# Core Python imports
import os
import pathlib
from glob import glob

# Third party imports
import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube

# punchbowl imports
from punchbowl.data import NormalizedMetadata
from punchbowl.data.meta import MetaField
from punchbowl.exceptions import InvalidDataError
from punchbowl.level3.f_corona_model import subtract_f_corona_background, subtract_f_corona_background_task

TEST_DIRECTORY = pathlib.Path(__file__).parent.resolve()
TESTDATA_DIR = os.path.dirname(__file__)
SAMPLE_FITS_PATH = os.path.join(TESTDATA_DIR, "L0_CL1_20211111070246_PUNCHData.fits")

@pytest.fixture()
def sample_data_list():
    number_elements = 25
    data_list = []
    for iStep in range(number_elements):
        data_list.append(SAMPLE_FITS_PATH)
    return data_list


@pytest.fixture()
def one_data(shape: tuple = (2048, 2048)) -> np.ndarray:
    """
    Generate some random data for testing
    """
    data = np.ones(shape)

    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"SECTION": {
        "TYPECODE": MetaField("TYPECODE", "", "CL", str, True, True, ""),
        "LEVEL": MetaField("LEVEL", "", "1", str, True, True, ""),
        "OBSRVTRY": MetaField("OBSRVTRY", "", "0", str, True, True, ""),
        "DATE-OBS": MetaField("DATE-OBS", "", "2008-01-03T04:57:00", str, True, True, "")}})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)

@pytest.fixture()
def observation_data(shape: tuple = (2048, 2048)) -> np.ndarray:
    """
    Generate some random data for testing
    """
    data = np.ones(shape)

    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"SECTION": {
        "TYPECODE": MetaField("TYPECODE", "", "CL", str, True, True, ""),
        "LEVEL": MetaField("LEVEL", "", "1", str, True, True, ""),
        "OBSRVTRY": MetaField("OBSRVTRY", "", "0", str, True, True, ""),
        "DATE-OBS": MetaField("DATE-OBS", "", "2008-01-03T08:57:00", str, True, True, "")}})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def zero_data(shape: tuple = (2048, 2048)) -> np.ndarray:
    """
    Generate some random data for testing
    """
    data = np.zeros(shape)

    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"SECTION": {
        "TYPECODE": MetaField("TYPECODE", "", "CL", str, True, True, ""),
        "LEVEL": MetaField("LEVEL", "", "1", str, True, True, ""),
        "OBSRVTRY": MetaField("OBSRVTRY", "", "0", str, True, True, ""),
        "DATE-OBS": MetaField("DATE-OBS", "", "2008-01-03T12:57:00", str, True, True, "")}})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def incorrect_shape_data(shape: tuple = (512, 512)) -> np.ndarray:
    """
    Generate some random data for testing
    """
    data = np.zeros(shape)

    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 256, 256
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape

    meta = NormalizedMetadata({"SECTION": {
        "TYPECODE": MetaField("TYPECODE", "", "CL", str, True, True, ""),
        "LEVEL": MetaField("LEVEL", "", "1", str, True, True, ""),
        "OBSRVTRY": MetaField("OBSRVTRY", "", "0", str, True, True, ""),
        "DATE-OBS": MetaField("DATE-OBS", "", "2008-01-03T08:57:00", str, True, True, "")}})

    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


def test_basic_subtraction(observation_data: NDCube, one_data: NDCube, zero_data: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    subtraction_punchdata = subtract_f_corona_background(observation_data, one_data, zero_data)
    assert isinstance(subtraction_punchdata, NDCube)
    assert np.all(subtraction_punchdata.data == 0.5)

def test_flipped_dates_subtraction_fails(observation_data: NDCube, one_data: NDCube, zero_data: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    with pytest.raises(InvalidDataError):
        _ = subtract_f_corona_background(observation_data, zero_data, one_data)


def test_after_is_before_subtraction_fails(observation_data: NDCube, one_data: NDCube, zero_data: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    with pytest.raises(InvalidDataError):
        _ = subtract_f_corona_background(observation_data, one_data, one_data)


def test_different_array_size_subtraction(incorrect_shape_data: NDCube, zero_data: NDCube) -> None:
    """
    dataset of increasing values passed in, a bad pixel map is passed in
    """
    with pytest.raises(Exception):
        subtract_f_corona_background_task(incorrect_shape_data, zero_data)



from datetime import UTC, datetime

import astropy.units as u
import numpy as np
from ndcube import NDCollection, NDCube
from prefect import get_run_logger
from solpolpy import resolve

from punchbowl.data.meta import NormalizedMetadata, set_spacecraft_location_to_earth
from punchbowl.prefect import punch_task


@punch_task
def convert_polarization(
        input_data: NDCube) -> NDCube:
    """Convert polarization from MZP to BpB."""
    logger = get_run_logger()
    logger.info("convert2bpb started")

    collection_contents = [(label,
                            NDCube(data=input_data[i].data,
                                   wcs=input_data.wcs.dropaxis(2),
                                   meta={"POLAR": angle}))
                           for (label, i, angle) in zip(["M", "Z", "P"],
                                                        [0, 1, 2],
                                                        [-60 * u.deg, 0 * u.deg, 60 * u.deg], strict=False)]
    data_collection = NDCollection(collection_contents, aligned_axes="all")

    resolved_data_collection = resolve(data_collection, "BpB", imax_effect=False)

    new_data = np.stack([resolved_data_collection["B"].data, resolved_data_collection["pB"].data], axis=0)
    new_wcs = input_data.wcs.copy()

    output_meta = NormalizedMetadata.load_template("PTM", "3")
    output_meta["DATE"] = datetime.now(UTC).isoformat()
    output_meta["DATE-AVG"] = input_data.meta["DATE-AVG"].value
    output_meta["DATE-OBS"] = input_data.meta["DATE-OBS"].value
    output_meta["DATE-BEG"] = input_data.meta["DATE-BEG"].value
    output_meta["DATE-END"] = input_data.meta["DATE-END"].value
    output = NDCube(data=new_data, wcs=new_wcs, meta=output_meta)
    output = set_spacecraft_location_to_earth(output)

    logger.info("convert2bpb finished")
    output.meta.history.add_now("LEVEL3-convert2bpb", "Convert MZP to BpB")

    return output



import os
import pathlib
from datetime import UTC, datetime, timedelta

import pytest
from ndcube import NDCube

from punchbowl.data.punch_io import write_ndcube_to_fits
from punchbowl.data.tests.test_punch_io import sample_ndcube
from punchbowl.level3.flow import level3_core_flow

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()

# We run these tests first because when they were at the end they timed out on Prefect's fixture.
# Putting them first fixes that. Not sure why.

@pytest.mark.order(1)
def test_core_flow_runs_with_filenames(sample_ndcube, tmpdir):
    data_list = [sample_ndcube(shape=(3, 10, 10), code="PTM", level="2") for _ in range(12)]
    before_f_corona_model_path = os.path.join(tmpdir, "before_f_corona.fits")
    before_f_corona_model = sample_ndcube(shape=(3, 10, 10), code="PFM", level="3")
    before_f_corona_model.meta['DATE-OBS'] = str(datetime(2024, 2, 22, 16, 0, 1) - timedelta(hours=5))
    write_ndcube_to_fits(before_f_corona_model, before_f_corona_model_path)

    after_f_corona_model_path = os.path.join(tmpdir, "after_f_corona.fits")
    after_f_corona_model = sample_ndcube(shape=(3, 10, 10), code="PFM", level="3")
    after_f_corona_model.meta['DATE-OBS'] = str(datetime(2024, 2, 22, 16, 0, 1) + timedelta(hours=5))
    write_ndcube_to_fits(after_f_corona_model, after_f_corona_model_path)

    paths = []
    for i, cube in enumerate(data_list):
        path = os.path.join(tmpdir, f"test_input_{i}.fits")
        write_ndcube_to_fits(cube, path)
        paths.append(path)

    output = level3_core_flow(data_list,
                              before_f_corona_model_path, after_f_corona_model_path,
                              None)
    assert isinstance(output[0], NDCube)

@pytest.mark.order(1)
def test_core_flow_clear_runs_with_filenames(sample_ndcube, tmpdir):
    data_list = [sample_ndcube(shape=(10, 10), code="CTM", level="2") for _ in range(12)]
    before_f_corona_model_path = os.path.join(tmpdir, "before_f_corona.fits")
    before_f_corona_model = sample_ndcube(shape=(10, 10), code="CFM", level="3")
    before_f_corona_model.meta['DATE-OBS'] = str(datetime(2024, 2, 22, 16, 0, 1) - timedelta(hours=5))
    write_ndcube_to_fits(before_f_corona_model, before_f_corona_model_path)

    after_f_corona_model_path = os.path.join(tmpdir, "after_f_corona.fits")
    after_f_corona_model = sample_ndcube(shape=(10, 10), code="CFM", level="3")
    after_f_corona_model.meta['DATE-OBS'] = str(datetime(2024, 2, 22, 16, 0, 1) + timedelta(hours=5))
    write_ndcube_to_fits(after_f_corona_model, after_f_corona_model_path)

    paths = []
    for i, cube in enumerate(data_list):
        path = os.path.join(tmpdir, f"test_input_{i}.fits")
        write_ndcube_to_fits(cube, path)
        paths.append(path)

    output = level3_core_flow(data_list,
                              before_f_corona_model_path, after_f_corona_model_path,
                              None)
    assert isinstance(output[0], NDCube)

@pytest.mark.order(1)
@pytest.mark.parametrize("num_files", [1, 3])
def test_core_flow_runs_with_objects_and_calibration_files(tmpdir, sample_ndcube, num_files):
    data_list = [sample_ndcube(shape=(3, 10, 10), code="PTM", level="2") for _ in range(num_files)]

    before_f_corona_model_path = os.path.join(tmpdir, "before_f_corona.fits")
    before_f_corona_model = sample_ndcube(shape=(3, 10, 10), code="PFM", level="3")
    before_f_corona_model.meta['DATE-OBS'] =  str(datetime(2024, 2, 22, 16, 0, 1) - timedelta(hours=5))
    write_ndcube_to_fits(before_f_corona_model, before_f_corona_model_path)

    after_f_corona_model_path = os.path.join(tmpdir, "after_f_corona.fits")
    after_f_corona_model = sample_ndcube(shape=(3, 10, 10), code="PFM", level="3")
    after_f_corona_model.meta['DATE-OBS'] = str(datetime(2024, 2, 22, 16, 0, 1) + timedelta(hours=5))
    write_ndcube_to_fits(after_f_corona_model, after_f_corona_model_path)

    output = level3_core_flow(data_list,
                              before_f_corona_model_path, after_f_corona_model_path,
                              None)
    assert isinstance(output[0], NDCube)



import os
import pathlib

from ndcube import NDCube

from punchbowl.data.punch_io import write_ndcube_to_fits
from punchbowl.data.tests.test_punch_io import sample_ndcube
from punchbowl.level1.flow import level1_core_flow

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()

def test_core_flow_runs_with_filenames(sample_ndcube, tmpdir):
        input_name = os.path.join(tmpdir, "test_input.fits")
        output_name = os.path.join(tmpdir, "test_output.fits")
        sample_data = sample_ndcube(shape=(10, 10), code="CR1", level="1")
        sample_data.meta["RAWBITS"] = 16
        sample_data.meta["COMPBITS"] = 10
        sample_data.meta["GAINLEFT"] = 4.9
        sample_data.meta["GAINRGHT"] = 4.9
        sample_data.meta["OFFSET"] = 100
        write_ndcube_to_fits(sample_data, input_name)

        quartic_coefficient_path = THIS_DIRECTORY / "data" / "test_quartic_coeffs.fits"
        vignetting_path = THIS_DIRECTORY / "data" / "PUNCH_L1_GR1_20240222163425_v1.fits"
        output = level1_core_flow([input_name],
                                  quartic_coefficient_path=quartic_coefficient_path,
                                  vignetting_function_path=vignetting_path,
                                  output_filename=[output_name])
        assert isinstance(output[0], NDCube)
        assert os.path.exists(output_name[0])



def test_core_flow_runs_with_objects_and_calibration_files(sample_ndcube):
    cube = sample_ndcube(shape=(10, 10), code="CR1", level="1")
    cube.meta["RAWBITS"] = 16
    cube.meta["COMPBITS"] = 10
    cube.meta["GAINLEFT"] = 4.9
    cube.meta["GAINRGHT"] = 4.9
    cube.meta["OFFSET"] = 100
    quartic_coefficient_path = THIS_DIRECTORY / "data" / "test_quartic_coeffs.fits"
    vignetting_path = THIS_DIRECTORY / "data" / "PUNCH_L1_GR1_20240222163425_v1.fits"

    output = level1_core_flow([cube],
                              quartic_coefficient_path=quartic_coefficient_path,
                              vignetting_function_path=vignetting_path,)
    assert isinstance(output[0], NDCube)



from datetime import UTC, datetime

import pytest
from pytest import fixture

from punchbowl.data.history import History, HistoryEntry
from punchbowl.data.meta import NormalizedMetadata


@fixture
def empty_history():
    return History()


def test_history_add_one(empty_history):
    entry = HistoryEntry(datetime.now(UTC), "test", "dummy")
    assert len(empty_history) == 0
    empty_history.add_entry(entry)
    assert len(empty_history) == 1
    assert empty_history.most_recent().source == "test"
    assert empty_history.most_recent().comment == "dummy"
    assert empty_history.most_recent() == empty_history[-1]
    empty_history.clear()
    assert len(empty_history) == 0


def test_history_iterate(empty_history):
    empty_history.add_entry(HistoryEntry(datetime.now(UTC), "0", "0"))
    empty_history.add_entry(HistoryEntry(datetime.now(UTC), "1", "1"))
    empty_history.add_entry(HistoryEntry(datetime.now(UTC), "2", "2"))

    for i, entry in enumerate(empty_history):
        assert entry.comment == str(i), "history objects not read in order"


def test_empty_history_equal():
    assert History() == History()


def test_history_equality():
    entry = HistoryEntry(datetime(2023, 10, 30, 12, 20), "test", "test comment")

    h1 = History()
    h1.add_entry(entry)

    h2 = History()
    h2.add_entry(entry)

    assert h1 == h2


def test_history_not_equals_if_different():
    h1 = History()
    h1.add_now("Test", "one")

    h2 = History()
    h2.add_now("Test", "two")

    assert h1 != h2


def test_history_cannot_compare_to_nonhistory_type():
    h1 = History()
    h2 = {"Not": "History"}
    with pytest.raises(TypeError):
        h1 == h2


def test_empty_history_from_fits_header():
    m = NormalizedMetadata.load_template("PM1", "0")
    m['DATE-OBS'] = "2024-01-01T00:00:00"
    m.delete_section("World Coordinate System")
    h = m.to_fits_header(wcs=None)

    assert History.from_fits_header(h) == History()


def test_filled_history_from_fits_header(tmpdir):
    constructed_history = History()
    constructed_history.add_entry(HistoryEntry(datetime(2023, 10, 30, 12, 20), "test", "test comment"))
    constructed_history.add_entry(HistoryEntry(datetime(2023, 10, 30, 12, 20), "test2", "test comment"))

    m = NormalizedMetadata.load_template("PM1", "0")
    m.history.add_entry(HistoryEntry(datetime(2023, 10, 30, 12, 20), "test", "test comment"))
    m.history.add_entry(HistoryEntry(datetime(2023, 10, 30, 12, 20), "test2", "test comment"))
    m.delete_section("World Coordinate System")
    h = m.to_fits_header()

    assert History.from_fits_header(h) == constructed_history



import os
import csv
from importlib.metadata import version as get_version

from packaging.version import Version

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "punchbowl"
copyright = "2025, PUNCH Science Operations Center"
author = "PUNCH Science Operations Center"

# The full version, including alpha/beta/rc tags
release: str = get_version("punchbowl")
version: str = release
_version = Version(release)
if _version.is_devrelease:
    version = release = f"{_version.base_version}.dev{_version.dev}"


# -- Metadata documentation ---------------------------------------------------

def copy_and_truncate_csv(src_path, dest_path, columns, filter_column, filter_value, default_key):
    with open(src_path, mode='r', newline='') as src_file:
        reader = csv.DictReader(src_file)
        filtered_rows = [
            {
                **{col: row[col] for col in columns},
                'COMMENT': row['DEFAULT'] if not row['COMMENT'] else row['COMMENT']
            } if row[filter_column] == filter_value else None
            for row in reader
        ]
        filtered_rows = [row for row in filtered_rows if row]  # Remove None entries

    os.makedirs(os.path.dirname(dest_path), exist_ok=True)
    with open(dest_path, mode='w', newline='') as dest_file:
        writer = csv.DictWriter(dest_file, fieldnames=columns)
        writer.writeheader()
        writer.writerows(filtered_rows)

filter_column = 'TYPE'
filter_value = 'keyword'

src_csv_path = os.path.abspath(os.path.join('../punchbowl/data/data/', 'omniheader.csv'))
dest_csv_path = os.path.abspath(os.path.join('./data/', 'omniheader_select.csv'))
columns_to_include = ['KEYWORD', 'COMMENT']

copy_and_truncate_csv(src_csv_path, dest_csv_path, columns_to_include, filter_column, filter_value, 'DEFAULT')


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = ["autoapi.extension",
              "sphinx.ext.autodoc",
              "sphinx.ext.napoleon",
              "sphinx_rtd_theme",
              "sphinx_favicon",
              'sphinx_gallery.gen_gallery',
              "sphinxcontrib.mermaid",
              # 'jupyterlite_sphinx',
              "IPython.sphinxext.ipython_console_highlighting"]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
html_show_sourcelink = False
html_static_path = ["_static"]
html_theme_options = {
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/punch-mission/punchbowl",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        }
    ],
    "show_nav_level": 1,
    "show_toc_level": 3,
    "logo": {
        "text": "punchbowl",
        "image_light": "_static/logo.png",
        "image_dark": "_static/logo.png",
    }
}
html_context = {
    # "github_url": "https://github.com", # or your GitHub Enterprise site
    "github_user": "punch-mission",
    "github_repo": "punchbowl",
    "github_version": "main",
    "doc_path": "docs/source/",
}


autoapi_dirs = ["../punchbowl"]

favicons = ["favicon.ico"]

mermaid_verbose = True

sphinx_gallery_conf = {
    'filename_pattern': '^((?!skip_).)*$',
    'examples_dirs': os.path.join('..', 'examples'),
    'gallery_dirs': 'auto_examples',
    'jupyterlite': {
        'use_jupyter_lab': True,
        'notebook_modification_function': None,
        'jupyterlite_contents': "jupyterlite_contents"
    },
    "binder": {
        "org": "punch-mission",
        "repo": "punchbowl",
        "branch": "binder",
        "binderhub_url": "https://mybinder.org",
        "dependencies": "./binder/requirements.txt",
        "use_jupyter_lab": True,
    },
}



import numpy as np
import pytest
from astropy.wcs import WCS
from ndcube import NDCube
from prefect.logging import disable_run_logger

from punchbowl.data.tests.test_punch_io import sample_ndcube
from punchbowl.level2.merge import merge_many_polarized_task


@pytest.fixture
def sample_data_list(sample_ndcube):
    """
    Generate a list of sample PUNCH data objects for testing
    """
    sample_pd1 = sample_ndcube((4096, 4096), code="PM1", level="1")
    sample_pd1.data[100:300, 300:400] = 1
    sample_pd1.data[500:700, 800:900] = 1
    sample_pd1.uncertainty.array[100:300, 300:400] = 1
    sample_pd1.uncertainty.array[500:700, 800:900] = 1

    sample_pd2 = sample_ndcube((4096, 4096), code="PM1", level="1")
    sample_pd2.data[100:300, 300:400] = 500
    sample_pd2.data[500:700, 800:900] = 500
    sample_pd2.uncertainty.array[100:300, 300:400] = np.inf
    sample_pd2.uncertainty.array[500:700, 800:900] = 0
    return [sample_pd1, sample_pd2]


def test_merge_many_task(sample_data_list):
    """
    Test the image_merge prefect flow using a test harness
    """
    trefoil_wcs = WCS("level2/data/trefoil_hdr.fits")
    trefoil_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
    trefoil_shape  = (4096, 4096)
    trefoil_wcs.array_shape = trefoil_shape

    with disable_run_logger():
        output_punchdata = merge_many_polarized_task.fn(sample_data_list, trefoil_wcs)
    assert isinstance(output_punchdata, NDCube)
    assert output_punchdata.data.shape == (3, 4096, 4096)
    assert np.allclose(output_punchdata.data[0, 100:300, 300:400], 1)
    assert np.allclose(output_punchdata.data[0, 500:700, 800:900], 1)
    assert np.allclose(output_punchdata.uncertainty.array[0, 100:300, 300:400], 1)
    assert np.allclose(output_punchdata.uncertainty.array[0, 500:700, 800:900], 1)



import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from skimage.color import lab2rgb


def _cmap_punch() -> LinearSegmentedColormap:
    """Generate PUNCH colormap."""
    # Define key colors in LAB space
    black_lab = np.array([0, 0, 0])
    orange_lab = np.array([50, 15, 50])
    white_lab = np.array([100, 0, 0])

    # Define the number of colors
    n = 256
    lab_colors = np.zeros((n, 3))

    # Transition from black to orange
    for i in range(n // 2):
        t = i / (n // 2 - 1)
        lab_colors[i] = black_lab * (1 - t) + orange_lab * t

    # Transition from orange to white
    for i in range(n // 2, n):
        t = (i - n // 2) / (n // 2 - 1)
        lab_colors[i] = orange_lab * (1 - t) + white_lab * t

    rgb_colors = lab2rgb(lab_colors.reshape(1, -1, 3)).reshape(n, 3)
    return LinearSegmentedColormap.from_list("PUNCH", rgb_colors, N=n)

cmap_punch = _cmap_punch()

def radial_distance(h: int, w: int, center: tuple[int, int] | None = None, radius: float | None = None) -> np.ndarray:
    """Create radial distance array."""
    if center is None:
        center = (int(w/2), int(h/2))

    if radius is None:
        radius = min([center[0], center[1], w-center[0], h-center[1]])

    y, x = np.ogrid[:h, :w]
    dist_arr = np.sqrt((x - center[0])**2 + (y - center[1])**2)

    return dist_arr / dist_arr.max()


def radial_filter(data: np.ndarray) -> np.ndarray:
    """Filter data with radial distance function."""
    return data * radial_distance(*data.shape) ** 2.5



import astropy
import numpy as np
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube

from punchbowl.data import NormalizedMetadata
from punchbowl.prefect import punch_task
from punchbowl.util import average_datetime


@punch_task
def merge_many_polarized_task(data: list[NDCube | None], trefoil_wcs: WCS) -> NDCube:
    """Merge many task and carefully combine uncertainties."""
    trefoil_data_layers, trefoil_uncertainty_layers = [], []
    for polarization in [-60, 0, 60]:
        selected_images = [d for d in data if d is not None and d.meta["POLAR"].value == polarization]
        if len(selected_images) > 0:
            reprojected_data = np.stack([d.data for d in selected_images], axis=-1)
            reprojected_uncertainties = np.stack([d.uncertainty.array for d in selected_images], axis=-1)
            reprojected_uncertainties[reprojected_uncertainties <= 0] = np.inf
            reprojected_uncertainties[np.isinf(reprojected_uncertainties)] = 1E64
            reprojected_uncertainties[reprojected_data <= 0] = np.inf

            reprojected_weights = 1 / np.square(reprojected_uncertainties)

            trefoil_data_layers.append(np.nansum(reprojected_data * reprojected_weights, axis=2) /
                                       np.nansum(reprojected_weights, axis=2))
            trefoil_uncertainty_layers.append(1/np.nansum(np.sqrt(reprojected_weights), axis=2))
        else:
            trefoil_data_layers.append(np.zeros((4096, 4096)))
            trefoil_uncertainty_layers.append(np.zeros((4096, 4096))-999)

    output_meta = NormalizedMetadata.load_template("PTM", "2")
    output_dateobs = average_datetime([d.meta.datetime for d in data if d is not None]).isoformat()
    output_datebeg = min([d.meta.datetime for d in data if d is not None]).isoformat()
    output_dateend = max([d.meta.datetime for d in data if d is not None]).isoformat()
    output_meta["DATE-OBS"] = output_dateobs
    output_meta["DATE-BEG"] = output_datebeg
    output_meta["DATE-END"] = output_dateend
    trefoil_3d_wcs = astropy.wcs.utils.add_stokes_axis_to_wcs(trefoil_wcs, 2)
    return NDCube(
        data=np.stack(trefoil_data_layers, axis=0),
        uncertainty=StdDevUncertainty(np.stack(trefoil_uncertainty_layers, axis=0)),
        wcs=trefoil_3d_wcs,
        meta=output_meta,
    )

@punch_task
def merge_many_clear_task(data: list[NDCube | None], trefoil_wcs: WCS, level: str = "2") -> NDCube:
    """Merge many task and carefully combine uncertainties."""
    trefoil_data_layers, trefoil_uncertainty_layers = [], []
    selected_images = [d for d in data if d is not None]

    if len(selected_images) > 0:
        reprojected_data = np.stack([d.data for d in selected_images], axis=-1)
        reprojected_uncertainties = np.stack([d.uncertainty.array for d in selected_images], axis=-1)
        reprojected_uncertainties[reprojected_uncertainties <= 0] = np.inf
        reprojected_uncertainties[np.isinf(reprojected_uncertainties)] = 1E64
        reprojected_uncertainties[reprojected_data <= 0] = np.inf

        reprojected_weights = 1/np.square(reprojected_uncertainties)

        trefoil_data_layers.append(np.nansum(reprojected_data * reprojected_weights, axis=-1) /
                                   np.nansum(reprojected_weights, axis=-1))
        trefoil_uncertainty_layers.append(1/np.sqrt(np.nansum(reprojected_weights, axis=-1)))
    else:
        trefoil_data_layers.append(np.zeros((4096, 4096)))
        trefoil_uncertainty_layers.append(np.zeros((4096, 4096))-999)

    output_meta = NormalizedMetadata.load_template("CTM", level=level)

    return NDCube(
        data=np.stack(trefoil_data_layers, axis=0).squeeze(),
        uncertainty=StdDevUncertainty(np.stack(trefoil_uncertainty_layers, axis=0).squeeze()),
        wcs=trefoil_wcs,
        meta=output_meta,
    )



import pathlib
from datetime import datetime

import pytest
from ndcube import NDCube
from prefect.logging import disable_run_logger

from punchbowl.data.tests.test_punch_io import sample_ndcube
from punchbowl.exceptions import (
    IncorrectPolarizationStateWarning,
    IncorrectTelescopeWarning,
    InvalidDataError,
    LargeTimeDeltaWarning,
)
from punchbowl.level1.vignette import correct_vignetting_task

THIS_DIRECTORY = pathlib.Path(__file__).parent.resolve()


@pytest.mark.prefect_test()
def test_check_calibration_time_delta_warning(sample_ndcube) -> None:
    """
    If the time between the data of interest and the calibration file is too great, then a warning is raised.
    """

    sample_data = sample_ndcube(shape=(10, 10), level="1")
    sample_data.meta['DATE-OBS'].value = str(datetime(2022, 2, 22, 16, 0, 1))
    vignetting_filename = THIS_DIRECTORY / "data" / "PUNCH_L1_GR1_20240222163425_v1.fits"

    with disable_run_logger():
        with pytest.warns(LargeTimeDeltaWarning):
            corrected_punchdata = correct_vignetting_task.fn(sample_data, vignetting_filename)
            assert isinstance(corrected_punchdata, NDCube)


@pytest.mark.prefect_test()
def test_no_vignetting_file(sample_ndcube) -> None:
    """
    An invalid vignetting file should be provided. Check that an error is raised.
    """

    sample_data = sample_ndcube(shape=(10, 10), code="CR1", level="0")
    vignetting_filename = None

    with disable_run_logger():
        corrected_punchdata = correct_vignetting_task.fn(sample_data, vignetting_filename)
        assert isinstance(corrected_punchdata, NDCube)
        assert corrected_punchdata.meta.history[0].comment == 'Vignetting skipped'


@pytest.mark.prefect_test()
def test_invalid_vignetting_file(sample_ndcube) -> None:
    """
    An invalid vignetting file should be provided. Check that an error is raised.
    """

    sample_data = sample_ndcube(shape=(10, 10), code="CR1", level="0")
    vignetting_filename = THIS_DIRECTORY / "data" / "bogus_filename.fits"

    with disable_run_logger():
        with pytest.raises(InvalidDataError):
            corrected_punchdata = correct_vignetting_task.fn(sample_data, vignetting_filename)


@pytest.mark.prefect_test()
def test_invalid_polarization_state(sample_ndcube) -> None:
    """
    Check that a mismatch between polarization states in the vignetting function and data raises an error.
    """

    sample_data = sample_ndcube(shape=(10, 10), level="1")
    vignetting_filename = THIS_DIRECTORY / "data" / "PUNCH_L1_GR1_20240222163425_v1.fits"

    with disable_run_logger():
        with pytest.warns(IncorrectPolarizationStateWarning):
            corrected_punchdata = correct_vignetting_task.fn(sample_data, vignetting_filename)
            assert isinstance(corrected_punchdata, NDCube)


@pytest.mark.prefect_test()
def test_invalid_telescope(sample_ndcube) -> None:
    """
    Check that a mismatch between telescopes in the vignetting function and data raises an error.
    """

    sample_data = sample_ndcube(shape=(10, 10), level="1")
    sample_data.meta['TELESCOP'].value = 'PUNCH-2'
    vignetting_filename = THIS_DIRECTORY / "data" / "PUNCH_L1_GR1_20240222163425_v1.fits"

    with disable_run_logger():
        with pytest.warns(IncorrectTelescopeWarning):
            corrected_punchdata = correct_vignetting_task.fn(sample_data, vignetting_filename)
            assert isinstance(corrected_punchdata, NDCube)


@pytest.mark.prefect_test()
def test_invalid_data_file(sample_ndcube) -> None:
    """
    An invalid vignetting file should be provided. Check that an error is raised.
    """

    sample_data = sample_ndcube(shape=(10, 10), code="CR1", level="0")
    vignetting_filename = THIS_DIRECTORY / "non_existent_file.fits"

    with disable_run_logger():
        with pytest.raises(InvalidDataError):
            corrected_punchdata = correct_vignetting_task.fn(sample_data, vignetting_filename)


@pytest.mark.prefect_test()
def test_vignetting_correction(sample_ndcube) -> None:
    """
    A valid vignetting file should be provided. Check that a corrected PUNCHData object is generated.
    """

    sample_data = sample_ndcube(shape=(10, 10), code="CR1", level="1")
    vignetting_filename = THIS_DIRECTORY / "data" / "PUNCH_L1_GR1_20240222163425_v1.fits"

    with disable_run_logger():
        corrected_punchdata = correct_vignetting_task.fn(sample_data, vignetting_filename)

    assert isinstance(corrected_punchdata, NDCube)



from __future__ import annotations

from datetime import UTC, datetime
from collections import namedtuple

from astropy.io.fits import Header
from dateutil.parser import parse as parse_datetime

HistoryEntry = namedtuple("HistoryEntry", "datetime, source, comment")


class History:
    """Representation of the history of edits done to a PUNCH data object."""

    def __init__(self) -> None:
        """Create blank history."""
        self._entries: list[HistoryEntry] = []

    def add_entry(self, entry: HistoryEntry) -> None:
        """
        Add an entry to the History log.

        Parameters
        ----------
        entry : HistoryEntry
            A HistoryEntry object to add to the History log

        Returns
        -------
        None

        """
        self._entries.append(entry)

    def add_now(self, source: str, comment: str) -> None:
        """
        Add a new history entry at the current time.

        Parameters
        ----------
        source : str
            what module of the code the history entry originates from
        comment : str
            a note of what the history comment means

        Returns
        -------
        None

        """
        self._entries.append(HistoryEntry(datetime.now(UTC), source, comment))

    def clear(self) -> None:
        """
        Clear all the history entries so the History is blank.

        Returns
        -------
        None

        """
        self._entries = []

    def __getitem__(self, index: int) -> HistoryEntry:
        """
        Given an index, return the requested HistoryEntry.

        Parameters
        ----------
        index : int
            numerical index of the history entry, increasing number typically indicates an older entry

        Returns
        -------
        HistoryEntry
            history at specified `index`

        """
        return self._entries[index]

    def most_recent(self) -> HistoryEntry:
        """
        Get the most recent HistoryEntry, i.e. the youngest.

        Returns
        -------
        HistoryEntry
            returns HistoryEntry that is the youngest

        """
        return self._entries[-1]

    def __len__(self) -> int:
        """Get length."""
        return len(self._entries)

    def __str__(self) -> str:
        """
        Format a string combining all the history entries.

        Returns
        -------
        str
            a combined record of the history entries

        """
        return "\n".join([f"{e.datetime:%Y-%m-%dT%H:%M:%S} => {e.source} => {e.comment}|" for e in self._entries])

    def __iter__(self) -> History:
        """Iterate."""
        self.current_index = 0
        return self

    def __next__(self) -> HistoryEntry:
        """Get next."""
        if self.current_index >= len(self):
            raise StopIteration
        entry = self._entries[self.current_index]
        self.current_index += 1
        return entry

    def __eq__(self, other: History) -> bool:
        """Check equality of two History objects."""
        if not isinstance(other, History):
            msg = f"Can only check equality between two history objects, found History and {type(other)}"
            raise TypeError(msg)
        return self._entries == other._entries

    @classmethod
    def from_fits_header(cls, head: Header) -> History:
        """
        Construct a history from a FITS header.

        Parameters
        ----------
        head : Header
            a FITS header to read from

        Returns
        -------
        History
            the history derived from a given FITS header

        """
        if "HISTORY" not in head:
            out = cls()
        else:
            out = cls()
            entries = "".join([h for h in head["HISTORY"][1:]]).split("|")[:-1]  # noqa: C416
            for entry in entries:
                dt, source, comment = entry.split(" => ")
                dt = parse_datetime(dt)
                out.add_entry(HistoryEntry(dt, source, comment))
        return out



"""
========================
Guide to QuickPUNCH Data
========================

A notebook guide to working with QuickPUNCH data in Python
"""

# %%
# Load libraries

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from ndcube import NDCube
from sunpy.map import Map

from punchbowl.data.sample import QUICKPUNCH_NQN, QUICKPUNCH_WQM

# %%
# Open the HDU list, and read out the appropriate data
# As the data is RICE compressed, the *second* HDU contains the main data frame
# The third HDU contains a corresponding uncertainty array

with fits.open(QUICKPUNCH_WQM) as hdul:
    print("WFI QuickPUNCH HDU List:")
    hdul.info()
    wfi_qp_data = hdul[1].data
    wfi_qp_header = hdul[1].header
    wfi_qp_uncertainty = hdul[2].data

with fits.open(QUICKPUNCH_NQN) as hdul:
    print("NFI QuickPUNCH HDU List:")
    hdul.info()
    nfi_qp_data = hdul[1].data
    nfi_qp_header = hdul[1].header
    nfi_qp_uncertainty = hdul[2].data

# %%
# The primary data arrays are stored as standard ndarrays
# The uncertainty data array has the dimensions as the primary data array
# Both the primary and uncertainty data arrays share the same header, contained in the primary HDU

print("WFI data array size:", wfi_qp_data.shape)
print("WFI uncertainty array size:", wfi_qp_uncertainty.shape)

# %%
# The corresponding headers can be queried as AstroPy header objects

wfi_qp_header["DATE-OBS"]

# %%
# The header information can be converted into an AstroPy WCS object

wfi_qp_data_wcs = WCS(wfi_qp_header)
nfi_qp_data_wcs = WCS(nfi_qp_header)

# %%
# Construct a SunPy Map object of out this data

wfi_qp_data_map = Map(wfi_qp_data, wfi_qp_header)
nfi_qp_data_map = Map(nfi_qp_data, nfi_qp_header)

# %%
# Display this SunPy Map object

wfi_qp_data_map

# %%
# Display this SunPy Map object

nfi_qp_data_map

# %%
# Construct an NDCube object out of this data

wfi_qp_data_ndcube = NDCube(wfi_qp_data, wcs=wfi_qp_data_wcs)
nfi_qp_data_ndcube = NDCube(nfi_qp_data, wcs=nfi_qp_data_wcs)

# %%
# Take a quick look at these NDCube objects

wfi_qp_data_ndcube, nfi_qp_data_ndcube

# %%
# Display this data in a regular plotting environment, using the associated WCS

plt.figure(figsize=(7.5, 7.5))
ax = plt.subplot(111, projection=wfi_qp_data_wcs)
plt.imshow(np.log(wfi_qp_data), cmap="Greys_r", vmin=-16, vmax=0)
lon, lat = ax.coords
lat.set_ticks(np.arange(-90, 90, 5) * u.degree)
lon.set_ticks(np.arange(-180, 180, 5) * u.degree)
lat.set_major_formatter("dd")
lon.set_major_formatter("dd")
ax.set_facecolor("black")
ax.coords.grid(color="white", alpha=.1)
plt.xlabel("Helioprojective longitude")
plt.ylabel("Helioprojective latitude")
plt.scatter(0, 0, s=240, color="k", transform=ax.get_transform("world"))
plt.title("QuickPUNCH Mosaic total brightness - " + wfi_qp_header["DATE-OBS"] + "UT")

# %%
# Display this data in a regular plotting environment, using the associated WCS

plt.figure(figsize=(7.5, 7.5))
ax = plt.subplot(111, projection=nfi_qp_data_wcs)
plt.imshow(np.log(nfi_qp_data), cmap="Greys_r", vmin=-16, vmax=0)
lon, lat = ax.coords
lat.set_ticks(np.arange(-90, 90, 5) * u.degree)
lon.set_ticks(np.arange(-180, 180, 5) * u.degree)
lat.set_major_formatter("dd")
lon.set_major_formatter("dd")
ax.set_facecolor("black")
ax.coords.grid(color="white", alpha=.1)
plt.xlabel("Helioprojective longitude")
plt.ylabel("Helioprojective latitude")
plt.scatter(0, 0, s=240, color="k", transform=ax.get_transform("world"))
plt.title("QuickPUNCH NFI total brightness - " + nfi_qp_header["DATE-OBS"] + "UT")

# %%
# Again noting that these files are compressed, additional keywords will be visible when viewing these FITS files outside of Python.
# These keywords relate to the compression implementation, and can be retrieved using astropy.io.fits, if needed, using the disable_image_compression keyword.

with fits.open(QUICKPUNCH_WQM, disable_image_compression=True) as hdul:
    header_compression = hdul[1].header

header_compression

# %%



import os
from datetime import UTC, datetime

import numpy as np
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from astropy.wcs.utils import add_stokes_axis_to_wcs
from ndcube import NDCube

from punchbowl.data import NormalizedMetadata, get_base_file_name, write_ndcube_to_fits
from punchbowl.data.meta import load_level_spec

LEVELS = ["0", "1", "2", "3", "L", "Q"]


def sample_ndcube(shape, code="PM1", level="0"):
    data = np.zeros(shape).astype(np.float32)
    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.1, 0.1
    wcs.wcs.crpix = 0, 0
    wcs.wcs.crval = 1, 1

    if level in ["2", "3"] and code[0] == "P":
        wcs = add_stokes_axis_to_wcs(wcs, 2)
    wcs.array_shape = shape

    meta = NormalizedMetadata.load_template(code, level)
    meta['DATE-OBS'] = str(datetime(2024, 1, 1, 0, 0, 0))
    meta['DATE-BEG'] = str(datetime(2024, 1, 1, 0, 0, 0))
    meta['DATE-END'] = str(datetime(2024, 1, 1, 0, 0, 0))
    meta['DATE-AVG'] = str(datetime(2024, 1, 1, 0, 0, 0))
    meta['FILEVRSN'] = "1"
    meta['POLARREF'] = "Instrument"
    meta['POLAROFF'] = 0.0
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


def construct_all_product_headers(directory, level, outpath):
    date_obs = datetime.now(UTC)
    level_path = os.path.join(directory, f"Level{level}.yaml")
    level_spec = load_level_spec(level_path)
    product_keys = list(level_spec["Products"].keys())
    # crafts = load_spacecraft_def().keys()
    if level in ["0", "1", "2"]:
        crafts = {'1': '', '2': '', '3': '', '4': ''}.keys()
        shape = (2048,2048)
    if level in ["2", "3", "Q", "L"]:
        crafts = {'M': '', 'N': ''}.keys()
        shape = (4096,4096)
    if level in ["Q", "L"]:
        crafts = {'M': '', 'N': ''}.keys()
        shape = (1024,1024)
    product_keys = sorted(list(set([pc.replace("?", craft) for craft in crafts for pc in product_keys])))
    for pc in product_keys:
        try:
            meta = NormalizedMetadata.load_template(pc, level)
        except Exception as e:
            assert False, f"failed to create {pc} for level {level} because: {e}"
        meta['DATE-OBS'] = str(datetime.now(UTC))

        sample_data = sample_ndcube(shape=shape, code=pc, level=level)

        filename = outpath + get_base_file_name(sample_data) + '.fits'

        print('Finished writing ' + filename)

        write_ndcube_to_fits(sample_data, filename=filename, write_hash=False)


if __name__ == "__main__":

    path_yaml = '/Users/clowder/work/punch/punchbowl/punchbowl/data/data/'
    path_output = '/Users/clowder/data/punch/metadata/'

    for level in LEVELS:
        construct_all_product_headers(path_yaml, level, path_output)

    print("Job's finished.")



import numpy as np
from ndcube import NDCube

from punchbowl.data.meta import NormalizedMetadata
from punchbowl.prefect import punch_task


@punch_task
def create_low_noise_task(inputs: list[NDCube]) -> NDCube:
    """Create a low noise image from a set of inputs."""
    num_images = np.nansum([np.logical_not(np.isnan(c.data)) for c in inputs], axis=0)
    new_data = np.nanmean([c.data for c in inputs], axis=0)
    combined_uncertainty = np.sqrt(np.nansum(np.square([c.uncertainty.array for c in inputs]), axis=0))
    final_uncertainty = combined_uncertainty / np.sqrt(num_images)

    new_code = inputs[0].meta.product_code[0] + "A" + inputs[0].meta.product_code[2]
    new_meta = NormalizedMetadata.load_template(new_code, "3")
    for old_key, old_value in inputs[0].meta.items():
        new_meta[old_key] = old_value
    return NDCube(data=new_data, uncertainty=final_uncertainty, wcs=inputs[0].wcs, meta=new_meta)



import numpy as np
import pytest
from astropy.time import Time
from astropy.wcs import WCS
from prefect.testing.utilities import prefect_test_harness

from punchbowl.data.tests.test_punch_io import sample_ndcube
from punchbowl.level2.resample import reproject_cube, reproject_many_flow


@pytest.fixture
def sample_punchdata_list(sample_ndcube):
    """
    Generate a list of sample PUNCH data objects for testing
    """
    sample_pd1 = sample_ndcube((50, 50))
    sample_pd2 = sample_ndcube((50, 50))
    return [sample_pd1, sample_pd2]


def test_reproject_many_flow(sample_punchdata_list):
    trefoil_wcs = WCS("level2/data/trefoil_hdr.fits")
    trefoil_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"  # TODO: figure out why this is necessary, seems like a bug
    trefoil_shape = (128, 128)
    trefoil_wcs.array_shape = trefoil_shape
    with prefect_test_harness():
        output = reproject_many_flow(sample_punchdata_list, trefoil_wcs, trefoil_shape)
    for result in output:
        assert result.data.shape == trefoil_shape
        assert result.uncertainty.array.shape == trefoil_shape



import numpy as np
import pytest
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from ndcube import NDCube

# punchbowl imports
from punchbowl.data import NormalizedMetadata
from punchbowl.level2.bright_structure import run_zspike


@pytest.fixture()
def sample_bad_pixel_map(shape: tuple = (2048, 2048), n_bad_pixels: int = 20) -> NDCube:
    """
    Generate some random data for testing
    """
    bad_pixel_map = np.ones(shape)

    x_coords = np.fix(np.random.random(n_bad_pixels) * shape[0]).astype(int)
    y_coords = np.fix(np.random.random(n_bad_pixels) * shape[1]).astype(int)

    bad_pixel_map[x_coords, y_coords] = 0

    bad_pixel_map = bad_pixel_map.astype(int)

    uncertainty = StdDevUncertainty(bad_pixel_map)

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape[-2:]

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=bad_pixel_map, uncertainty=uncertainty, wcs=wcs, meta=meta)

@pytest.fixture()
def sample_punchdata(shape: tuple = (5, 2048, 2048)) -> NDCube:
    """
    Generate a sample PUNCH data object for testing
    """

    data = np.random.random(shape)
    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape[-2:]

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def even_sample_punchdata(shape: tuple = (6, 2048, 2048)) -> NDCube:
    """
    Generate a sample PUNCH data object for testing
    """

    data = np.random.random(shape)
    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape[-2:]

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def sample_zero_punchdata(shape: tuple = (5, 2048, 2048)) -> NDCube:
    """
    Generate a sample PUNCH data object for testing
    """

    data = np.zeros(shape)
    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape[-2:]

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


@pytest.fixture()
def one_bright_point_sample_punchdata(shape: tuple = (7, 2048, 2048)) -> NDCube:
    """
    Generate a sample PUNCH data object for testing
    """
    x_interest = 200
    y_interest = 200
    data = np.random.random(shape)

    # add a bright point
    data[3, x_interest, y_interest]=1000
    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))


    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape[-2:]

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)

@pytest.fixture()
def two_bright_point_sample_punchdata(shape: tuple = (7, 2048, 2048)) -> NDCube:
    """
    Generate a sample PUNCH data object for testing
    """
    x_interest = 200
    y_interest = 200
    data = np.random.random(shape)

    # add two bright points in a row
    data[3, x_interest, y_interest]=1000
    data[4, x_interest, y_interest]=1000
    uncertainty = StdDevUncertainty(np.sqrt(np.abs(data)))


    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-AZP", "HPLT-AZP"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.02, 0.02
    wcs.wcs.crpix = 1024, 1024
    wcs.wcs.crval = 0, 24.75
    wcs.array_shape = shape[-2:]

    meta = NormalizedMetadata({"TYPECODE": "CL", "LEVEL": "1", "OBSRVTRY": "0", "DATE-OBS": "2008-01-03 08:57:00"})
    return NDCube(data=data, uncertainty=uncertainty, wcs=wcs, meta=meta)


def test_valid_data_and_uncertainty(sample_punchdata: NDCube):
    with pytest.raises(Exception):
        # Call find_spikes with valid data and uncertainty
        result = run_zspike(sample_punchdata.data, sample_punchdata.uncertainty.array)
        # Perform assertions on the result
        # For example, check the shape or the datatype of the result
        assert isinstance(result, np.ndarray)
        assert result.shape == sample_punchdata.data.shape

def test_zero_threshold(sample_punchdata: NDCube):
    # test the thresholds are zero
    threshold = 0
    result = run_zspike(sample_punchdata.data,
                        sample_punchdata.uncertainty.array,
                        threshold=threshold,
                        diff_method='abs')
    assert np.sum(result) == 0

    assert np.sum(result) == 0


def test_diff_methods(sample_zero_punchdata: NDCube):
    result_abs = run_zspike(sample_zero_punchdata.data, sample_zero_punchdata.uncertainty.array, diff_method='abs')
    result_sigma = run_zspike(sample_zero_punchdata.data, sample_zero_punchdata.uncertainty.array, diff_method='sigma')
    assert np.array_equal(result_abs, result_sigma)


def test_different_parameters(sample_punchdata: NDCube):
    required_yes = 1
    veto_limit = 1
    dilation = 0
    result = run_zspike(sample_punchdata.data, sample_punchdata.uncertainty.array, required_yes=required_yes, veto_limit=veto_limit, dilation=dilation)
    assert result.shape == np.shape(sample_punchdata.data[0,:,:])

#####
def test_raise_error_insufficient_frames(sample_bad_pixel_map: NDCube):
    # creates a raise error as only a 2d array is passed in

    with pytest.raises(ValueError):
        #sample_data.write(SAMPLE_WRITE_PATH)
        run_zspike(sample_bad_pixel_map.data,
                   sample_bad_pixel_map.uncertainty.array)


def test_raise_error(even_sample_punchdata: NDCube):
    # creates a raise error as an even array is passed in

    with pytest.raises(ValueError):
        #sample_data.write(SAMPLE_WRITE_PATH)
        run_zspike(even_sample_punchdata.data,
                   even_sample_punchdata.uncertainty.array)


def test_raise_no_error(even_sample_punchdata: NDCube):
    # does not create a raise error as an even array is passed in and an index of interest

    result=run_zspike(even_sample_punchdata.data,
                      even_sample_punchdata.uncertainty.array,
                      index_of_interest=3)

    assert isinstance(result, np.ndarray)
    assert result.shape == np.shape(even_sample_punchdata.data[0,:,:])

def test_single_bright_point(sample_punchdata: NDCube):
    # test passes with single bright point
    test_data=sample_punchdata.data
    test_uncertainty=sample_punchdata.uncertainty.array
    test_data.data[3, 200, 200]=1000
    # add spike

    result=run_zspike(test_data, test_uncertainty)

    assert result.shape == np.shape(sample_punchdata.data[0,:,:])
    assert isinstance(result, np.ndarray)

def test_single_bright_point_2(one_bright_point_sample_punchdata: NDCube):

    x_interest = 200
    y_interest = 200

    result_2 = run_zspike(one_bright_point_sample_punchdata.data,
                          one_bright_point_sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1)

    # test cell of interest is set to 'True'
    assert bool(result_2[x_interest, y_interest]) is True

    # test other cells are set to 'False'
    assert bool(result_2[x_interest+1, y_interest]) is False


def test_veto(two_bright_point_sample_punchdata: NDCube):
    # test works with one vote
    x_interest = 200
    y_interest = 200

    result_2 = run_zspike(two_bright_point_sample_punchdata.data,
                          two_bright_point_sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1)

    # test cell of interest is set to 'False' with veto
    assert bool(result_2[x_interest, y_interest]) is True

    two_bright_point_sample_punchdata.uncertainty.array[:, :, :] = 0


    result_3 = run_zspike(two_bright_point_sample_punchdata.data,
                          two_bright_point_sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=0,
                          index_of_interest=-1)

    # test cell of interest is set to 'False' with veto
    assert bool(result_3[x_interest, y_interest]) is False


def test_uncertainty(sample_punchdata: NDCube):
    # create an uncertainty array of 0's
    sample_punchdata.uncertainty.array[...] = 0

    # choose a pixel of interest
    x_test_px=210
    y_test_px=355
    index_of_interest=-1

   # initial test to see pixel of interest is normal
    result_0 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    # test cell of interest is set to 'True'
    assert bool(result_0[y_test_px, x_test_px]) is False

    # set pixel of interest to high value
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px] = 1000

    result_1 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    # test cell of interest is set to 'True'
    assert bool(result_1[y_test_px, x_test_px]) is True

    # make bad pixels adjacent to cell of interest also high
    # set pixel of interest to high value
    sample_punchdata.data[:, y_test_px, x_test_px]=1000

    result_2 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    # test cell of interest is set to 'False'
    assert bool(result_2[y_test_px, x_test_px]) is False

    # set surrounding values to uncertain

    sample_punchdata.uncertainty.array[index_of_interest, y_test_px, x_test_px] = np.inf
    result_3 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)
    # test cell of interest is set to 'True' due to uncertainty flag set on the surrounding pixels
    assert bool(result_3[y_test_px, x_test_px]) is True




def test_threshold_abs(sample_punchdata: NDCube):
    # create an uncertainty array of 0's
    sample_punchdata.uncertainty.array[:, :, :] = 0

    # choose a pixel of interest
    x_test_px=210
    y_test_px=355
    index_of_interest=-1

   # initial test to see pixel of interest is normal
    result_0 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)


    assert bool(result_0[y_test_px, x_test_px]) is False

    # set pixel of interest to high value
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=100

    result_1 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_1[y_test_px, x_test_px]) is True

    # make bad pixel threshold high

    result_2 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=300,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    # test cell of interest is set to 'False'
    assert bool(result_2[y_test_px, x_test_px]) is False


def test_threshold_sigma(sample_punchdata: NDCube):
    # create an uncertainty array of 0's
    sample_punchdata.uncertainty.array[:, :, :] = 0
    sample_punchdata.data[:, :, :] = 0

    # choose a pixel of interest
    x_test_px=210
    y_test_px=355
    index_of_interest=-1

    # initial test to see pixel of interest is normal
    result_0 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=3,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)


    assert bool(result_0[y_test_px, x_test_px]) is False

    # set pixel of interest to high value
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=100


    result_1 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_1[y_test_px, x_test_px]) is True

    # make bad pixel threshold high

    result_2 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=300,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    # test cell of interest is set to 'False'
    assert bool(result_2[y_test_px, x_test_px]) is False



def test_required_yes_abs(sample_punchdata: NDCube):
    # create an uncertainty array of 0's
    sample_punchdata.uncertainty.array[:, :, :] = 0

    # choose a pixel of interest
    x_test_px=210
    y_test_px=355
    index_of_interest=-1


    # set pixel of interest to high value

    result = run_zspike(sample_punchdata.data,
                        sample_punchdata.uncertainty.array,
                        diff_method='abs',
                        threshold=1,
                        required_yes=1,
                        veto_limit=1,
                        index_of_interest=index_of_interest)

    assert bool(result[y_test_px, x_test_px]) is False

    #sample_punchdata.data[0:3, y_test_px, x_test_px]=100
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=100

    # set pixel of interest to high value

    result_1 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_1[y_test_px, x_test_px]) is True

    # change the number adjacent elements that have high values, this
    # reduces the number of available yes voters, making the cell of
    # interest 'false'
    sample_punchdata.data[0:1, y_test_px, x_test_px]=100
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=100

    # set pixel of interest to high value

    result_2 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_2[y_test_px, x_test_px]) is True

    # change the number adjacent elements that have high values
    sample_punchdata.data[0:1, y_test_px, x_test_px]=100
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=100

    # set pixel of interest to high value

    result_3 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=1,
                          required_yes=3,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_3[y_test_px, x_test_px]) is False


def test_required_yes_sigma(sample_punchdata: NDCube):
    # create an uncertainty and data array of 0's
    sample_punchdata.uncertainty.array[:, :, :] = 0
    sample_punchdata.data[:, :, :] = 0
    # choose a pixel of interest
    x_test_px=210
    y_test_px=355
    index_of_interest=-1


    result = run_zspike(sample_punchdata.data,
                        sample_punchdata.uncertainty.array,
                        diff_method='sigma',
                        threshold=1,
                        required_yes=1,
                        veto_limit=1,
                        index_of_interest=index_of_interest)

    assert bool(result[y_test_px, x_test_px]) is False

    #sample_punchdata.data[0:3, y_test_px, x_test_px]=100
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10

    # set pixel of interest to high value

    result_1 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_1[y_test_px, x_test_px]) is True

    # change the number adjacent elements that have high values, this
    # reduces the number of available yes voters, making the cell of
    # interest 'false'
    sample_punchdata.data[0:1, y_test_px, x_test_px]=10
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10

    # set pixel of interest to high value

    result_2 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_2[y_test_px, x_test_px]) is True

    # change the number adjacent elements that have high values
    sample_punchdata.data[0:1, y_test_px, x_test_px]=10
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10

    # set pixel of interest to high value

    result_3 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=3,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_3[y_test_px, x_test_px]) is False

def test_dilation_abs(sample_punchdata: NDCube):
        # create an uncertainty and data array of 0's
    sample_punchdata.uncertainty.array[:, :, :] = 0
    #sample_punchdata.data[:, :, :] = 0
    # choose a pixel of interest
    x_test_px=210
    y_test_px=355
    index_of_interest=-1


    result = run_zspike(sample_punchdata.data,
                        sample_punchdata.uncertainty.array,
                        diff_method='abs',
                        threshold=1,
                        required_yes=1,
                        veto_limit=1,
                        index_of_interest=index_of_interest)

    assert bool(result[y_test_px, x_test_px]) is False

    #sample_punchdata.data[0:3, y_test_px, x_test_px]=100
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10

    # set pixel of interest to high value

    result_1 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_1[y_test_px, x_test_px]) is True

    # change the number adjacent elements that have high values, this
    # reduces the number of available yes voters, making the cell of
    # interest 'false'
    sample_punchdata.data[0:1, y_test_px, x_test_px]=100
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=100

    # set pixel of interest to high value

    result_2 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='abs',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)


    assert bool(result_2[y_test_px, x_test_px]) is True
    # with no dilation a an adjacent pixel is false
    assert bool(result_2[y_test_px+1, x_test_px]) is False

    result_3 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest,
                          dilation=10)


    assert bool(result_3[y_test_px, x_test_px]) is True
    # with dilation an adjacent pixel is true
    assert bool(result_3[y_test_px+1, x_test_px]) is True



    # change the number adjacent elements that have high values
    sample_punchdata.data[0:1, y_test_px, x_test_px]=10
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10



def test_dilation_sigma(sample_punchdata: NDCube):
        # create an uncertainty and data array of 0's
    sample_punchdata.uncertainty.array[:, :, :] = 0
    sample_punchdata.data[:, :, :] = 0
    # choose a pixel of interest
    x_test_px=210
    y_test_px=355
    index_of_interest=-1


    result = run_zspike(sample_punchdata.data,
                        sample_punchdata.uncertainty.array,
                        diff_method='sigma',
                        threshold=1,
                        required_yes=1,
                        veto_limit=1,
                        index_of_interest=index_of_interest)

    assert bool(result[y_test_px, x_test_px]) is False

    #sample_punchdata.data[0:3, y_test_px, x_test_px]=100
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10

    # set pixel of interest to high value

    result_1 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)

    assert bool(result_1[y_test_px, x_test_px]) is True

    # change the number adjacent elements that have high values, this
    # reduces the number of available yes voters, making the cell of
    # interest 'false'
    sample_punchdata.data[0:1, y_test_px, x_test_px]=10
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10

    # set pixel of interest to high value

    result_2 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest)


    assert bool(result_2[y_test_px, x_test_px]) is True
    # with no dilation a an adjacent pixel is false
    assert bool(result_2[y_test_px+1, x_test_px]) is False

    result_3 = run_zspike(sample_punchdata.data,
                          sample_punchdata.uncertainty.array,
                          diff_method='sigma',
                          threshold=1,
                          required_yes=1,
                          veto_limit=1,
                          index_of_interest=index_of_interest,
                          dilation=10)


    assert bool(result_3[y_test_px, x_test_px]) is True
    # with dilation an adjacent pixel is true
    assert bool(result_3[y_test_px+1, x_test_px]) is True



    # change the number adjacent elements that have high values
    sample_punchdata.data[0:1, y_test_px, x_test_px]=10
    sample_punchdata.data[index_of_interest, y_test_px, x_test_px]=10



import numpy as np
from scipy.fftpack import fftn, fftshift, ifftn, ifftshift
from skimage.filters import window
from sunpy import log

try:
    import cupy
    HAS_CUPY = True
except ImportError:
    HAS_CUPY = False


def layer_mask(radius: float, img_shape: (int, int)) -> np.ndarray:
    """
    Generate a circular mask centered in the output array.

    Parameters
    ----------
    radius : float
        The radius of the masked circular region
    img_shape : tuple
        The shape of the output mask

    Returns
    -------
    np.ndarray
        Circular mask

    """
    xs = np.arange(0, img_shape[1]) - img_shape[1]/2 + .5
    ys = np.arange(0, img_shape[0]) - img_shape[0]/2 + .5
    x2s, y2s = np.meshgrid(xs**2, ys**2)
    distance = np.sqrt(x2s + y2s)
    return distance < radius


def generate_hourglass_filter(fft_cube: np.ndarray, cutoff_velocity: float) -> np.ndarray:
    """
    Create an hourglass filter mask.

    Parameters
    ----------
    fft_cube : np.ndarray
        The 3D Fourier space cube.
    cutoff_velocity : float
        The cutoff velocity.

    Returns
    -------
    np.ndarray
        Hourglass filter mask.

    """
    fft_shape = fft_cube.shape
    img_shape = (fft_shape[1], fft_shape[2])

    mask1 = np.stack(
        [layer_mask(radius, img_shape)
         for radius in np.linspace(int(fft_shape[1] / 2), cutoff_velocity, int(fft_shape[0] / 2))],
        axis=0)
    mask2 = np.stack(
        [layer_mask(radius, img_shape)
         for radius in np.linspace(cutoff_velocity, int(fft_shape[1] / 2), int(fft_shape[0] / 2))],
        axis=0)

    return np.concatenate((mask1, mask2), axis=0)


def apply_motion_filter(stacked_data:np.ndarray,
                  apod_margin: int,
                  use_gpu: bool = True) -> np.ndarray:
    """
    Perform a Fourier motion filter on input datacube.

    Parameters
    ----------
    stacked_data : np.ndarray
        Starfield removed stacked datacube.
    apod_margin : int
        Apodization margin.
    use_gpu : bool, optional
        Whether to use GPU for processing (default is True).

    Returns
    -------
    np.ndarray
        Fourier motion filtered datacube.

    """
    padded_data = np.pad(stacked_data, ((apod_margin * 2, apod_margin * 2), (apod_margin, apod_margin),
                                        (apod_margin, apod_margin)), mode="constant", constant_values=0)

    h = window(("tukey", 0.1), padded_data[0].shape)  # Creating a 2D window
    wimage = padded_data * h

    if use_gpu and not HAS_CUPY:
        log.info("cupy not installed or working, falling back to CPU")
    if HAS_CUPY and use_gpu:
        wimage = cupy.ndarray(wimage)

    # Performing 3D FFT
    fwfft = fftshift(fftn(wimage))
    filt_mask = generate_hourglass_filter(fwfft, 100)
    filt_fft = filt_mask * fwfft  # Adding hourglass mask to Forward FFT

    # Performing inverse FFT
    inv_shift = ifftshift(filt_fft)
    invfft = ifftn(inv_shift)

    return cupy.asnumpy(invfft) if (HAS_CUPY and use_gpu) else invfft



import astropy.units as u
import numpy as np
from astropy.wcs import WCS

from punchbowl.data import units
from punchbowl.data.tests.test_punch_io import sample_ndcube


def test_calculate_image_pixel_area_rotated_input():
    # When we rotate the imager, the on-sky areas of the pixels shouldn't change
    rotations = [0, 30, 45, 60, 90, 120] * u.deg
    shape = (512, 512)
    area_maps = []
    for rotation in rotations:
        wcs = WCS(naxis=2)
        wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
        wcs.wcs.pc = np.array([[np.cos(rotation), -np.sin(rotation)], [np.sin(rotation), np.cos(rotation)]])
        wcs.wcs.crval = 0, 0
        wcs.wcs.crpix = 256, 256
        wcs.wcs.cdelt = 0.05, 0.05
        wcs.array_shape = shape

        area_maps.append(units.calculate_image_pixel_area(wcs, shape))

    for area_map in area_maps[1:]:
        np.testing.assert_allclose(area_maps[0], area_map)


def test_calculate_image_pixel_area_output_shape():
    # Test that the output data has the right shape, meaning axes probably aren't being transposed
    shape = (960, 1024)
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
    wcs.wcs.pc = np.eye(2)
    wcs.wcs.crval = 0, 0
    wcs.wcs.crpix = 512, 480
    wcs.wcs.cdelt = 0.05, 0.05
    wcs.array_shape = shape

    area_map = units.calculate_image_pixel_area(wcs, shape)
    assert area_map.shape == shape

def test_split_ccd_array():
    shape = (10,10)
    values = (1,2)
    array = units.split_ccd_array(shape, values[0], values[1])
    assert array.shape == shape
    assert array.mean() == (values[0] + values[1]) / 2

def test_dn_to_msb(sample_ndcube):
    shape = (10,10)
    data_dn = sample_ndcube(shape)

    data_dn.data[...] = (np.random.random(shape) * 10).astype(int)
    data_msb = units.dn_to_msb(data_dn.data, data_wcs=data_dn.wcs)
    assert data_msb.shape == shape



import copy
from collections.abc import Callable

import numpy as np
from astropy.wcs import WCS
from ndcube import NDCube
from thuban.pointing import refine_pointing

from punchbowl.data.wcs import calculate_celestial_wcs_from_helio, calculate_helio_wcs_from_celestial
from punchbowl.prefect import punch_task


@punch_task
def align_task(data_object: NDCube, mask: Callable | None = None) -> NDCube:
    """
    Determine the pointing of the image and updates the metadata appropriately.

    Parameters
    ----------
    data_object : NDCube
        data object to align
    mask : Callable | None
        function accepting coordinates and returning them only if they are not masked out

    Returns
    -------
    NDCube
        a modified version of the input with the WCS more accurately determined

    """
    saved_wcs = copy.deepcopy(data_object.wcs)
    celestial_input = calculate_celestial_wcs_from_helio(copy.deepcopy(data_object.wcs),
                                                         data_object.meta.astropy_time,
                                                         data_object.data.shape)
    refining_data = data_object.data.copy()
    refining_data[np.isinf(refining_data)] = 0
    refining_data[np.isnan(refining_data)] = 0
    celestial_output, observed_coords, result, trial_num = refine_pointing(refining_data, celestial_input,
                                                                        num_stars=10,
                                                                        detection_threshold=10,
                                                                        background_height=8, background_width=8,
                                                                        max_trials=32,
                                                                        dimmest_magnitude=6.0,
                                                                        chisqr_threshold=0.1,
                                                                        sigma=0.1,
                                                                        edge=100,
                                                                        mask=mask,
                                                                        method="least_squares")

    recovered_wcs, _ = calculate_helio_wcs_from_celestial(celestial_output,
                                                       data_object.meta.astropy_time,
                                                       data_object.data.shape)
    recovered_wcs.cpdis1 = data_object.wcs.cpdis1
    recovered_wcs.cpdis2 = data_object.wcs.cpdis2

    saved_wcs_fits = saved_wcs.to_fits()
    saved_header = saved_wcs_fits[0].header
    recovered_header = recovered_wcs.to_fits()[0].header
    for key in recovered_header:
        saved_header[key] = recovered_header[key]
    recovered_wcs = WCS(saved_header, saved_wcs_fits)

    output = NDCube(data=data_object.data,
                    wcs=recovered_wcs,
                    uncertainty=data_object.uncertainty,
                    unit=data_object.unit,
                    meta=data_object.meta)
    output.meta.history.add_now("LEVEL1-Align", "alignment done")
    return output



from __future__ import annotations

import os
import string
import hashlib
import os.path
import subprocess
from typing import Any
from pathlib import Path

import astropy.units as u
import lxml.etree as et
import numpy as np
from astropy.io import fits
from astropy.io.fits import Header
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from glymur import Jp2k, jp2box
from matplotlib.colors import LogNorm
from ndcube import NDCube
from PIL import Image, ImageDraw, ImageFont

from punchbowl.data.meta import NormalizedMetadata
from punchbowl.data.visualize import cmap_punch

_ROOT = os.path.abspath(os.path.dirname(__file__))

CALIBRATION_ANNOTATION = "{OBSRVTRY} - {TYPECODE}{OBSCODE} - {DATE-OBS} - exptime: {EXPTIME} s - polarizer: {POLAR} deg"


def write_file_hash(path: str) -> None:
    """Create a SHA-256 hash for a file."""
    file_hash = hashlib.sha256()
    with open(path, "rb") as f:
        fb = f.read()
        file_hash.update(fb)

    with open(path + ".sha", "w") as f:
        f.write(file_hash.hexdigest())


def get_base_file_name(cube: NDCube) -> str:
    """Determine the base file name without file type extension."""
    obscode = cube.meta["OBSCODE"].value
    file_level = cube.meta["LEVEL"].value
    type_code = cube.meta["TYPECODE"].value
    date_string = cube.meta.datetime.strftime("%Y%m%d%H%M%S")
    file_version = cube.meta["FILEVRSN"].value
    file_version = "?" if file_version == "" else file_version  # file version should never be empty!
    return "PUNCH_L" + file_level + "_" + type_code + obscode + "_" + date_string + "_v" + file_version


class DefaultFormatter(string.Formatter):
    """A formatter that doesn't fail if a keyword is missing. Used for quicklook."""

    def get_field(self, field_name: str, args: Any, kwargs: Any) -> str:
        """Provide a special getter that returns the name if it fails."""
        try:
            return super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError, IndexError):
            return "{" + field_name + "}", ()


def _header_to_xml(header: Header) -> et.Element:
    """
    Convert image header metadata into an XML Tree that can be inserted into a JPEG2000 file header.

    (Helper function adapted from SunPy)
    """
    fits = et.Element("fits")
    already_added = set()
    for key in header:
        if (key in already_added):
            continue
        already_added.add(key)
        el = et.SubElement(fits, key)
        data = header.get(key)
        data = ("1" if data else "0") if isinstance(data, bool) else str(data)
        el.text = data
    return fits


def _generate_jp2_xmlbox(header: Header) -> jp2box.XMLBox:
    """
    Generate the JPEG2000 XML box to be inserted into the JPEG2000 file.

    (Helper function adapted from SunPy)
    """
    header_xml = _header_to_xml(header)
    meta = et.Element("meta")
    meta.append(header_xml)
    tree = et.ElementTree(meta)
    return jp2box.XMLBox(xml=tree)


def write_ndcube_to_quicklook(cube: NDCube,
                              filename: str,
                              layer: int | None = None,
                              vmin: float = 1e-15,
                              vmax: float = 8e-12,
                              annotation: str | None = None,
                              color: bool = False) -> None:
    """
    Write an NDCube to a Quicklook format as a jpeg.

    Parameters
    ----------
    cube : NDCube
        data cube to visualize
    filename : str
        path to save output, must end in .jp2, .j2k, .jpeg, .jpg
    layer : int | None
        if the cube is 3D, then selects cube.data[layer] for visualization
    vmin : float
        the lower limit value to visualize
    vmax : float
        the upper limit value to visualize
    annotation : str | None
        a formatted string to add to the bottom of the image as a label
        can access metadata by key, e.g. "typecode={TYPECODE}" would write the data's typecode into the image
    color : bool
        flag to generate RGB quicklook files, grayscale by default

    Returns
    -------
    None

    """
    if (len(cube.data.shape) != 2) and layer is None:
        msg = "Output data must be two-dimensional, or a layer must be specified"
        raise ValueError(msg)

    if not filename.endswith((".jp2", ".j2k", ".jpeg", ".jpg")):
        msg = ("Filename must have a valid file extension `.jpeg`, `jpg`, `.jp2` or `.j2k`"
               f"Found: {os.path.splitext(filename)[1]}")
        raise ValueError(msg)

    norm = LogNorm(vmin=vmin, vmax=vmax)

    if layer is not None:  # noqa: SIM108
        image = cube.data[layer, :, :]
    else:
        image = cube.data

    if color:
        mode = "RGBA"
        scaled_arr = (cmap_punch(norm(np.flipud(image))) * 255).astype(np.uint8)
        fill_value = (255, 255, 255)
    else:
        mode = "L"
        scaled_arr = (np.clip(norm(np.flipud(image)), 0, 1) * 255).astype(np.uint8)
        fill_value = 255

    pil_image = Image.fromarray(scaled_arr, mode=mode)

    if annotation:
        pad_height = int(image.shape[1] * 50 / 2048)
        padded_image = Image.new(mode, (pil_image.width, pil_image.height + pad_height))

        padded_image.paste(pil_image, (0, 0))

        draw = ImageDraw.Draw(padded_image)
        font = ImageFont.load_default(size=int(pad_height / 2))

        formatter = DefaultFormatter()
        text = formatter.format(annotation, **cube.meta)
        text_offset = int(10 * image.shape[1] / 2048)
        text_position = (text_offset, pil_image.height + text_offset)
        draw.text(text_position, text, font=font, fill=fill_value)
        pil_image = padded_image

    arr_image = np.array(pil_image)

    tmp_filename = f"{filename}tmp.jp2"
    jp2 = Jp2k(tmp_filename, arr_image)
    meta_boxes = jp2.box
    target_index = len(meta_boxes) - 1
    header = cube.meta.to_fits_header(wcs=cube.wcs)
    header.remove("COMMENT", ignore_missing=True, remove_all=True)
    fits_box = _generate_jp2_xmlbox(header)
    meta_boxes.insert(target_index, fits_box)
    jp2.wrap(filename, boxes=meta_boxes)
    os.remove(tmp_filename)


def write_quicklook_to_mp4(files: list[str],
                           filename: str,
                           framerate: int = 5,
                           resolution: int = 1024,
                           codec: str = "libx264",
                           ffmpeg_cmd: str = "ffmpeg",
                           ) -> None:
    """
    Write a list of input quicklook jpeg2000 files to an output mp4 animation.

    Parameters
    ----------
    files : list[str]
        List of input files to animate
    filename : str
        Output filename
    framerate : int, optional
        Frame rate (default 5)
    resolution : int, optional
        Output resolution (default 1024)
    codec : str, optional
        Codec to use for encoding. For GPU acceleration.
        "h264_videotoolbox" can be used on ARM Macs, "h264_nvenc" can be used on Intel machines.
    ffmpeg_cmd : str
        path to the ffmpeg executable

    """
    input_sequence = f"concat:{'|'.join(files)}"

    ffmpeg_command = [
        ffmpeg_cmd,
        "-framerate", str(framerate),
        "-i", input_sequence,
        "-vf", f"scale=-1:{resolution}",
        "-c:v", codec,
        "-pix_fmt", "yuv420p",
        "-y",
        filename,
    ]

    subprocess.run(ffmpeg_command, check=False)  # noqa: S603


def write_ndcube_to_fits(cube: NDCube,
                         filename: str,
                         overwrite: bool = False,
                         write_hash: bool = True,
                         skip_stats: bool = False,
                         uncertainty_quantize_level: float = 16) -> None:
    """Write an NDCube as a FITS file."""
    if not filename.endswith(".fits"):
        msg = (
            "Filename must have a valid file extension `.fits`"
            f"Found: {os.path.splitext(filename)[1]}"
        )
        raise ValueError(
            msg,
        )

    if not skip_stats:
        _update_statistics(cube)

    full_header = cube.meta.to_fits_header(wcs=cube.wcs)
    full_header["FILENAME"] = os.path.basename(filename)

    hdu_data = fits.CompImageHDU(data=cube.data,
                                 header=full_header,
                                 name="Primary data array")
    hdu_provenance = fits.BinTableHDU.from_columns(fits.ColDefs([fits.Column(
        name="provenance", format="A40", array=np.char.array(cube.meta.provenance))]))
    hdu_provenance.name = "File provenance"

    hdul = cube.wcs.to_fits()
    hdul[0] = fits.PrimaryHDU()
    hdul.insert(1, hdu_data)
    if cube.meta["LEVEL"].value != "0":
        hdu_uncertainty = fits.CompImageHDU(data=_pack_uncertainty(cube),
                                            header=full_header,
                                            name="Uncertainty array",
                                            quantize_level=uncertainty_quantize_level)
        hdul.insert(2, hdu_uncertainty)
    hdul.append(hdu_provenance)
    hdul.writeto(filename, overwrite=overwrite, checksum=True)
    hdul.close()
    if write_hash:
        write_file_hash(filename)



def _pack_uncertainty(cube: NDCube) -> np.ndarray:
    """Compress the uncertainty for writing to file."""
    return np.zeros_like(cube.data) - 999 if cube.uncertainty is None else 1 / (cube.uncertainty.array / cube.data)


def _unpack_uncertainty(uncertainty_array: np.ndarray, data_array: np.ndarray) -> np.ndarray:
    """Uncompress the uncertainty when reading from a file."""
    # This is (1/uncertainty_array) * data_array, but this way we save time on memory allocation
    with np.errstate(divide="ignore", invalid="ignore"):
        np.divide(1, uncertainty_array, out=uncertainty_array)
        np.multiply(data_array, uncertainty_array, out=uncertainty_array)
    return uncertainty_array


def _update_statistics(cube: NDCube) -> None:
    """Update image statistics in metadata before writing to file."""
    # TODO - Determine DSATVAL omniheader value in calibrated units for L1+

    cube.meta["DATAZER"] = len(np.where(cube.data == 0)[0])

    cube.meta["DATASAT"] = len(np.where(cube.data >= cube.meta["DSATVAL"].value)[0])

    cube.data[np.logical_not(np.isfinite(cube.data))] = 0
    nonzero_data = cube.data[np.isfinite(cube.data) * (cube.data != 0)].flatten()

    if len(nonzero_data) > 0:
        cube.meta["DATAAVG"] = np.nanmean(nonzero_data).item()
        cube.meta["DATAMDN"] = np.nanmedian(nonzero_data).item()
        cube.meta["DATASIG"] = np.nanstd(nonzero_data).item()
    else:
        cube.meta["DATAAVG"] = -999.0
        cube.meta["DATAMDN"] = -999.0
        cube.meta["DATASIG"] = -999.0

    percentile_percentages = [1, 10, 25, 50, 75, 90, 95, 98, 99]
    percentile_values = np.nanpercentile(nonzero_data, percentile_percentages)
    if np.any(np.isnan(percentile_values)):  # report nan if any of the values are nan
        percentile_values = [-999.0 for _ in percentile_percentages]

    for percent, value in zip(percentile_percentages, percentile_values, strict=False):
        cube.meta[f"DATAP{percent:02d}"] = value

    cube.meta["DATAMIN"] = float(np.nanmin(cube.data))
    cube.meta["DATAMAX"] = float(np.nanmax(cube.data))


def load_ndcube_from_fits(path: str | Path, key: str = " ", include_provenance: bool = True) -> NDCube:
    """Load an NDCube from a FITS file."""
    with fits.open(path) as hdul:
        primary_hdu = hdul[1]
        data = primary_hdu.data
        header = primary_hdu.header
        # Reset checksum and datasum to match astropy.io.fits behavior
        header["CHECKSUM"] = ""
        header["DATASUM"] = ""
        meta = NormalizedMetadata.from_fits_header(header)
        if include_provenance:
            if isinstance(hdul[-1], fits.hdu.BinTableHDU):
                meta._provenance = hdul[-1].data["provenance"]  # noqa: SLF001
            else:
                msg = "Provenance HDU does not appear to be BinTableHDU type."
                raise ValueError(msg)
        wcs = WCS(header, hdul, key=key)
        unit = u.ct

        if len(hdul) >= 3 and isinstance(hdul[2], fits.hdu.CompImageHDU):
            secondary_hdu = hdul[2]
            uncertainty = _unpack_uncertainty(secondary_hdu.data.astype(float), data)
            uncertainty = StdDevUncertainty(uncertainty)
        else:
            uncertainty = None

    return NDCube(
        data.view(dtype=data.dtype.newbyteorder()).byteswap().astype(float),
        wcs=wcs,
        uncertainty=uncertainty,
        meta=meta,
        unit=unit,
    )



import numpy as np

from punchbowl.level1.despike import radial_array, spikejones


def test_create_radial_array():
    shape = (7, 7)
    center = (2, 2)
    arr = radial_array(shape, center=center)
    expected = np.array([
             [2, 2, 2, 2, 2, 3, 4],
             [2, 1, 1, 1, 2, 3, 4],
             [2, 1, 0, 1, 2, 3, 4],
             [2, 1, 1, 1, 2, 3, 4],
             [2, 2, 2, 2, 2, 3, 4],
             [3, 3, 3, 3, 3, 4, 5],
             [4, 4, 4, 4, 4, 5, 5]
            ])
    assert isinstance(arr, np.ndarray)
    assert np.all(arr == expected)


def test_spikejones_with_one_spike():
    image = np.zeros((100, 100)) + 0.1
    image[50, 50] = 100
    image[49, 50] = 10

    image[30, 70] = 5
    output, spikes = spikejones(image)

    assert output.shape == image.shape
    assert np.isclose(output[50, 50], 0.1)
    assert np.isclose(output[49, 50], 0.1)
    assert np.isclose(output[30, 70], 0.1)



"""
===================
Guide to Data
===================

A notebook guide to working with PUNCH FITS files in Python
"""

# %%
# This Jupyter Notebook presents a comprehensive guide to the analysis and visualization of PUNCH FITS files using Python. The guide leverages a suite of Python libraries, including `matplotlib` for data visualization, numpy for numerical operations, `astropy` for FITS file handling and WCS manipulation, `sunpy` for solar data analysis, and `ndcube` for multi-dimensional data handling. The notebook serves as a robust resource for researchers and scientists, providing them with the tools and knowledge to effectively manipulate and interpret PUNCH FITS files. The guide's step-by-step approach, coupled with code snippets and detailed explanations, ensures a user-friendly experience, promoting the accessibility of PUNCH data for a wider scientific community.
#
# To install these dependencies, use `pip install -r requirements.txt` in your terminal.

# %%
# Load libraries

from copy import deepcopy

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from matplotlib.colors import LogNorm
from ndcube import NDCube
from sunpy.map import Map

from punchbowl.data.sample import PUNCH_PAM

# %%
# Data loading
# ------------
#
# We first specify the file we wish to load. For this notebook, let's use the included L3 PAM file. Then, we'll load it two different ways:
#
# 1. With `astropy.io`
# 2. With `SunPy Maps`
#
# The data are RICE compressed, so we'll have to look in the *second* at index 1 (since we're zero-indexed) HDU for the main data frame.
# There is also an uncertainty layer in the last HDU that we'll explore later.

# %%
with fits.open(PUNCH_PAM) as hdul:
    data = hdul[1].data
    header = hdul[1].header
    uncertainty = hdul[2].data

# %%
# First, let's look at the shape of the data. For this data product, total brightness and polarized brightness are stacked along the first dimension.
# The uncertainty array corresponds on a pixel-to-pixel basis with the data array.

# %%
print(data.shape, uncertainty.shape)

# %%
# We can look in the header for details about the data. A few things to note:
#
# - The header is divided into helpful sections such as "temporal information" and "instrument and spacecraft state." This aids in navigating the information. These sections are consistent for other products (with the addition of more sections depending on the product).
# - There are two world coordinate systems (WCS), one for the helio system and one for a celestial right ascension and declination system. They match each other and can be used interchangeably depending on the type of analysis.
# - The first dimension iterates over total brightness and polarized brightness, in that order. This is indicated by the `OBSLAYR1` and `OBSLAYR2` keywords.

# %%
print(header)

# %%
# We can create `Astropy WCS` objects from the header. Since the data are three dimensional, the WCS will be three dimensions. Luckily, we can slice WCSes.

# %%
data_wcs = WCS(header)
print(data_wcs)

# %%
# We can also access the celestial WCS by setting `key='A'`. "

# %%
data_wcs_celestial = WCS(header, key="A")
print(data_wcs_celestial)

# %%
# We can also use SunPy Maps to load the data. However, since Maps only support 2D images, it will warn us that it's truncating a dimension of the data and only showing the total brightness. We get a list of Maps where the first entry is the data and the second is the uncertainty.

# %%
data_map = Map(PUNCH_PAM)
data_map[0].plot()
plt.show()

# %%
# If you wish to make a map of the polarized brightness, we can also do that manually.

# %%
data_map = Map(data[1], header)
data_map.plot()
plt.show()

# %%
# Plotting data
# -------------

# %% [markdown]
# There are many ways to plot the data. We are working on more visualization tools that will be available with the release of the `punchbowl` package. First, we can plot the SunPy map.

# %%
data_map.plot(norm="log")
plt.show()

# %% [markdown]
# We can also use Matplotlib directly to modify the plot however we want. Notice that we set the projection so it is coordinate aware. Remember we have to slice the WCS as `data_wcs[0]` instead of `data_wcs` because our data are 3D but we want to view only the spatial dimensions.

# %%
# Plot the data using matplotlib manually
fig, ax = plt.subplots(figsize=(9.5, 7.5), subplot_kw={"projection":data_wcs[0]})

im = ax.imshow(data[1], cmap="Greys_r", norm=LogNorm(vmin=1.77e-15, vmax=3.7e-11))

# set up the axis labels
lon, lat = ax.coords
lat.set_ticks(np.arange(-90, 90, 15) * u.degree)
lon.set_ticks(np.arange(-180, 180, 15) * u.degree)
lat.set_major_formatter("dd")
lon.set_major_formatter("dd")
ax.set_facecolor("black")
ax.coords.grid(color="white", alpha=.25, ls="dotted")
ax.set_xlabel("Helioprojective longitude")
ax.set_ylabel("Helioprojective latitude")
ax.set_title("Synthetic PUNCH Polarized Brightness")
fig.colorbar(im, ax=ax, label="Mean Solar Brightness")
plt.show()

# %% [markdown]
# Since we set the projection, we could also use the celestial WCS instead.

# %%
# Plot the data using matplotlib manually, using the celestial frame

fig, ax = plt.subplots(figsize=(9.5, 7.5), subplot_kw={"projection":data_wcs_celestial[0]})

im = ax.imshow(data[0], cmap="Greys_r", norm=LogNorm(vmin=1.77e-15, vmax=3.7e-11))

# set up the axis labels
lon, lat = ax.coords
lat.set_ticks(np.arange(-90, 90, 15) * u.degree)
lon.set_ticks(np.arange(-180, 180, 15) * u.degree)
lat.set_major_formatter("dd")
lon.set_major_formatter("dd")
ax.set_facecolor("black")
ax.coords.grid(color="white", alpha=.25, ls="dotted")
ax.set_xlabel("RA")
ax.set_ylabel("DEC")
ax.set_title("Synthetic PUNCH Polarized Brightness")
fig.colorbar(im, ax=ax, label="Mean Solar Brightness")
plt.show()

# %% [markdown]
# Reprojecting the data
# ---------------------

# %% [markdown]
# What if we want want to manipulate the images? We can use built-in tools of SunPy Maps. For example, we can reproject.

# %%
# What if we wanted to reproject the data to a new arbitrary coordinate frame?
# First, define a new target WCS
new_wcs = deepcopy(data_map.wcs)
new_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
new_wcs.wcs.cunit = "deg", "deg"
new_wcs.array_shape = 1024, 1024
new_wcs.wcs.crpix = 512, 512
new_wcs.wcs.crval = 10, 10
new_wcs.wcs.cdelt = 0.0225, 0.0225
new_wcs.wcs.pc = (0.66,-0.33), (0.33,0.66)

# We can then reproject using SunPy maps
new_map = data_map.reproject_to(new_wcs)
new_map.plot()
plt.show()

# %% [markdown]
# Using NDCube
# ------------
#
# So far, we've only talked about directly manipulating the data or using SunPy Maps. However, we encourage you to utilize NDCubes. We use them in the data reduction pipeline too. [NDCube](https://docs.sunpy.org/projects/ndcube/en/stable/introduction.html) is a SunPy package for generalized n-dimensional data. We can use our manually loaded data to make an NDCube with all of the data in one place. There's even a way to bundle uncertainty directly into the cube.

# %%
data_ndcube = NDCube(data, wcs=data_wcs, meta=header, uncertainty=uncertainty)
print(data_ndcube)

# %% [markdown]
# We can then use all the great NDCube functionality! Much like SunPy Maps we can reproject even. Note that as of writing this guide, the uncertainty is dropped when NDCube reprojects data.

# %%
# make a target WCS
new_wcs = deepcopy(data_wcs)
new_wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC", "STOKES"
new_wcs.wcs.cunit = "deg", "deg", ""
new_wcs.array_shape = 2, 1024, 1024
new_wcs.wcs.crpix = 0, 0, 0
new_wcs.wcs.crval = 0, 0, 0
new_wcs.wcs.cdelt = 0.0225, 0.0225, 1.0

new_cube = data_ndcube.reproject_to(new_wcs)
new_cube.plot()
plt.show()

# %% [markdown]
# NDCubes have a plotter on them too. Notice we even get an interactive slider to manipulate the polarization axis.

# %%
new_cube.plot(interpolation="None", norm=LogNorm(vmin=1.77E-15, vmax=3.7E-11), cmap="Greys_r")
plt.show()

# %% [markdown]
# We encourage you to explore NDCube more. There are other helpful things like [cropping](https://docs.sunpy.org/projects/ndcube/en/stable/generated/gallery/slicing_ndcube.html#cropping-cube-using-world-coordinate-values-using-ndcube-ndcube-crop) that you can do.

# %% [markdown]
# Uncertainty
# -----------

# %% [markdown]
# Level 1, 2, and 3 PUNCH products come with an uncertainty estimate. This uncertainty is expressed as the fractional uncertainty ranging from 0 to 1. To get the absolute uncertainty, simply multiply the fractional uncertainty by the data value.
#
# There's an extra step though. Uncertainty is encoded as an integer in the FITS file to save space. When we inspect the uncertainty, we see it ranges from 0 to 255. Here, 255 corresponds to a fractional uncertainty of 1.0 or complete uncertainty.

# %%
print(uncertainty.min(), uncertainty.max())  # values range from 0 to 255

absolute_uncertainty = uncertainty.astype(float) / 255 * data
print(absolute_uncertainty.min(), absolute_uncertainty.max())  # now the uncertainty is in the same units as the data

# %% [markdown]
# Let's plot the data and uncertainty to get a feel for it. As mentioned earlier, analysis and visualization tools that make this easier will be released with the `punchbowl` package.

# %%
vmin, vmax = 1.77E-15, 3.7E-11

fig, axs = plt.subplots(figsize=(14, 6), ncols=2, subplot_kw={"projection":data_wcs[0]})

im = axs[0].imshow(data[0], norm=LogNorm(vmin=vmin, vmax=vmax), interpolation="None", cmap="Greys_r")
axs[0].set_title("Data")

axs[1].imshow(absolute_uncertainty[0], norm=LogNorm(vmin=vmin, vmax=vmax), interpolation="None", cmap="Greys_r")
axs[1].set_title("Uncertainty")

fig.colorbar(im, ax=axs, label="Mean Solar Brightness")

# set up the axis labels
for ax in axs:
    lon, lat = ax.coords
    lat.set_ticks(np.arange(-90, 90, 15) * u.degree)
    lon.set_ticks(np.arange(-180, 180, 15) * u.degree)
    lat.set_major_formatter("dd")
    lon.set_major_formatter("dd")
    ax.set_facecolor("black")
    ax.coords.grid(color="white", alpha=.25, ls="dotted")
    ax.set_xlabel("Helioprojective longitude")
    ax.set_ylabel("Helioprojective latitude")
plt.show()



from datetime import UTC, datetime

import numpy as np
from dateutil.parser import parse as parse_datetime_str
from ndcube import NDCube
from prefect import get_run_logger

from punchbowl.data import NormalizedMetadata, load_ndcube_from_fits
from punchbowl.data.wcs import load_quickpunch_mosaic_wcs
from punchbowl.level3.f_corona_model import fill_nans_with_interpolation, model_fcorona_for_cube
from punchbowl.prefect import punch_flow


@punch_flow(log_prints=True)
def construct_qp_f_corona_model(filenames: list[str], smooth_level: float = 3.0,
                                       reference_time: str | None = None) -> list[NDCube]:
    """Construct QuickPUNCH F corona model."""
    logger = get_run_logger()

    if reference_time is None:
        reference_time = datetime.now(UTC)
    elif isinstance(reference_time, str):
        reference_time = parse_datetime_str(reference_time)

    trefoil_wcs, trefoil_shape = load_quickpunch_mosaic_wcs()

    logger.info("construct_f_corona_background started")

    if len(filenames) == 0:
        msg = "Require at least one input file"
        raise ValueError(msg)

    filenames.sort()

    data_shape = trefoil_shape

    number_of_data_frames = len(filenames)
    data_cube = np.empty((number_of_data_frames, *data_shape), dtype=float)
    uncertainty_cube = np.empty((number_of_data_frames, *data_shape), dtype=float)

    meta_list = []
    obs_times = []

    logger.info("beginning data loading")
    for i, address_out in enumerate(filenames):
        data_object = load_ndcube_from_fits(address_out)
        data_cube[i, ...] = data_object.data
        uncertainty_cube[i, ...] = data_object.uncertainty.array
        obs_times.append(data_object.meta.datetime.timestamp())
        meta_list.append(data_object.meta)
    logger.info("ending data loading")

    reference_xt = reference_time.timestamp()
    model_fcorona, _ = model_fcorona_for_cube(obs_times, reference_xt, data_cube, clip_factor=smooth_level)
    model_fcorona[model_fcorona<=0] = np.nan
    model_fcorona = fill_nans_with_interpolation(model_fcorona)

    meta = NormalizedMetadata.load_template("CFM", "Q")
    meta["DATE-OBS"] = str(reference_time)
    output_cube = NDCube(data=model_fcorona.squeeze(),
                                meta=meta,
                                wcs=trefoil_wcs)

    return [output_cube]



import os
from datetime import UTC, datetime
from collections import Counter

import pytest
from astropy.io import fits
from astropy.wcs import WCS

from punchbowl.data.history import HistoryEntry
from punchbowl.data.meta import MetaField, NormalizedMetadata, construct_all_product_codes, load_spacecraft_def

TESTDATA_DIR = os.path.dirname(__file__)
SAMPLE_FITS_PATH_UNCOMPRESSED = os.path.join(TESTDATA_DIR, "test_data.fits")
SAMPLE_FITS_PATH_COMPRESSED = os.path.join(TESTDATA_DIR, "test_data.fits")
SAMPLE_WRITE_PATH = os.path.join(TESTDATA_DIR, "write_test.fits")
SAMPLE_OMNIBUS_PATH = os.path.join(TESTDATA_DIR, "omniheader.csv")
SAMPLE_LEVEL_PATH = os.path.join(TESTDATA_DIR, "LevelTest.yaml")
SAMPLE_SPACECRAFT_DEF_PATH = os.path.join(TESTDATA_DIR, "spacecraft.yaml")


def test_metafield_creation_keyword_too_long():
    """ cannot create an invalid metafield"""
    with pytest.raises(ValueError):
        MetaField("TOO LONG KEYWORD",
                  "What's up?", 3, int, False, True, -99)


def test_metafield_creation_kinds_do_not_match():
    """the value, default, and kind must all agree"""
    with pytest.raises(TypeError):
        MetaField("HI",
                  "What's up?", 3, str, False, True, "hi there")

    with pytest.raises(TypeError):
        MetaField("TOO LONG KEYWORD",
                  "What's up?", "hi there", str, False, True, -99)


def test_metafield_update():
    """ you can update a metafield if it is set to be mutable"""
    mf = MetaField("HI", "What's up?", 3, int, False, True, -99)
    assert mf.keyword == "HI"
    assert mf.comment == "What's up?"
    assert mf.value == 3
    assert mf._datatype is int
    assert not mf.nullable
    assert mf._mutable
    assert mf.default == -99

    mf.value = 100
    assert mf.value == 100

    mf.default = -999
    assert mf.default == -999


def test_metafield_not_mutable():
    """ you cannot update immutable metafields"""
    mf = MetaField("HI", "What's up?", 3, int, False, False, -99)
    assert mf.keyword == "HI"
    assert mf.comment == "What's up?"
    assert mf.value == 3
    assert mf._datatype is int
    assert not mf.nullable
    assert not mf._mutable
    assert mf.default == -99

    with pytest.raises(RuntimeError):
        mf.value = 100


def test_metafield_wrong_kind_for_update():
    """ you must abide by the kind for an update"""
    mf = MetaField("HI", "What's up?", 3, int, False, True, -99)
    assert mf.keyword == "HI"
    assert mf.comment == "What's up?"
    assert mf.value == 3
    assert mf._datatype is int
    assert not mf.nullable
    assert mf._mutable
    assert mf.default == -99

    with pytest.raises(TypeError):
        mf.value = "this is invalid"

    with pytest.raises(TypeError):
        mf.default = "this is invalid"


def test_normalizedmetadata_from_template_abq():
    result = NormalizedMetadata.load_template("ABT",
                                              omniheader_path=SAMPLE_OMNIBUS_PATH,
                                              level_spec_path=SAMPLE_LEVEL_PATH,
                                              spacecraft_def_path=SAMPLE_SPACECRAFT_DEF_PATH)
    assert isinstance(result, NormalizedMetadata)
    assert 'Section 1' in result.sections
    assert 'Section 2' in result.sections
    assert 'Section 3' in result.sections
    assert 'Section 4' in result.sections
    assert 'Section 5' not in result.sections
    assert 'Section 6' not in result.sections

    assert 'KEYOMIT1' not in result
    assert 'KEYOMIT2' not in result
    assert result['KEYALTER'].value == 2
    assert int(result['KEYALTER']) == 2
    assert result['OTHERKEY'].value == 'Test'
    with pytest.raises(TypeError):
        int(result['OTHERKEY'])
    assert result['TITLE'].value == 'My name is test-0'


def test_normalizedmetadata_from_template_cb1():
    result = NormalizedMetadata.load_template("CB1",
                                              omniheader_path=SAMPLE_OMNIBUS_PATH,
                                              level_spec_path=SAMPLE_LEVEL_PATH,
                                              spacecraft_def_path=SAMPLE_SPACECRAFT_DEF_PATH)
    assert isinstance(result, NormalizedMetadata)
    assert 'Section 1' in result.sections
    assert 'Section 2' in result.sections
    assert 'Section 3' in result.sections
    assert 'Section 4' in result.sections
    assert 'Section 5' not in result.sections

    assert 'KEYOMIT1' not in result
    assert 'KEYOMIT2' not in result
    assert result['KEYALTER'].value == 3
    assert result['OTHERKEY'].value == 'No test'
    assert result['TITLE'].value == 'A default value'
    assert result['TITLE'].comment == 'with a default comment'


def test_normalizedmetadata_to_fits_header(tmpdir):
    result = NormalizedMetadata.load_template("CB1",
                                              omniheader_path=SAMPLE_OMNIBUS_PATH,
                                              level_spec_path=SAMPLE_LEVEL_PATH,
                                              spacecraft_def_path=SAMPLE_SPACECRAFT_DEF_PATH)
    header = result.to_fits_header()
    header.tofile(os.path.join(str(tmpdir), "test.fits"), overwrite=True)
    assert isinstance(header, fits.Header)


def test_normalizedmetadata_get_keys():
    result = NormalizedMetadata.load_template("CB1",
                                              omniheader_path=SAMPLE_OMNIBUS_PATH,
                                              level_spec_path=SAMPLE_LEVEL_PATH,
                                              spacecraft_def_path=SAMPLE_SPACECRAFT_DEF_PATH)
    keys = result.fits_keys
    assert len(keys) == 5


def test_create_level0_normalized_metadata():
    m = NormalizedMetadata.load_template("PM1", "0")
    assert 'DATE-OBS' in m
    assert "OBSRVTRY" in m


def test_create_level3_normalized_metadata():
    m = NormalizedMetadata.load_template("PTM", "3")
    assert "GEOD_LAT" in m
    assert "HGLN_OBS" in m
    assert "HEEX_OBS" in m


def test_normalized_metadata_to_fits_writes_history():
    m = NormalizedMetadata.load_template("PM1", "0")
    m.history.add_now("Test", "does it write?")
    m.history.add_now("Test", "how about twice?")
    m.delete_section("World Coordinate System")
    h = m.to_fits_header()
    assert "HISTORY" in h
    assert "does it write?" in str(h['HISTORY'])
    assert "how about twice" in str(h['HISTORY'])


def test_normalizedmetadata_from_fits_header():
    m = NormalizedMetadata.load_template("PM1", "0")
    m['DESCRPTN'] = 'This is a test!'
    m.history.add_entry(HistoryEntry(datetime(2023, 10, 30, 12, 20), "test", "test comment"))
    m.history.add_entry(HistoryEntry(datetime(2023, 10, 30, 12, 20), "test2", "test comment"))
    m.delete_section("World Coordinate System")
    h = m.to_fits_header()

    recovered = NormalizedMetadata.from_fits_header(h)
    recovered.delete_section("World Coordinate System")

    assert recovered['HISTORY'] == m['HISTORY']


def test_load_spacecraft_yaml():
    """ tests that spacecraft yaml can be loaded and is well-formed"""
    sc_yaml = load_spacecraft_def(path=SAMPLE_SPACECRAFT_DEF_PATH)
    for i in range(1, 5):
        assert str(i) in sc_yaml
    assert "M" in sc_yaml
    assert "N" in sc_yaml
    assert "T" in sc_yaml

    for _, v in sc_yaml.items():
        assert "craftname" in v
        assert "crafttype" in v
        assert "obsname" in v


def test_construct_all_product_codes():
    codes = construct_all_product_codes(level="0")
    assert isinstance(codes, list)
    assert len(codes) == 36


def test_fits_header_compliance():
    code="PM1"
    level="0"

    wcs = WCS(naxis=2)
    wcs.wcs.ctype = "HPLN-ARC", "HPLT-ARC"
    wcs.wcs.cunit = "deg", "deg"
    wcs.wcs.cdelt = 0.1, 0.1
    wcs.wcs.crpix = 0, 0
    wcs.wcs.crval = 1, 1
    wcs.wcs.cname = "HPC lon", "HPC lat"

    m = NormalizedMetadata.load_template(code, level)

    m["DATE-OBS"] = str(datetime.now(UTC))
    m["DATE-END"] = str(datetime.now(UTC))

    h = m.to_fits_header(wcs = wcs)

    assert h[0] == 'T'

    allowed_duplicates = {'COMMENT', 'HISTORY'}
    keyword_counts = Counter(key for key in h.keys() if key not in allowed_duplicates)
    keyword_duplicates = [keyword for keyword, count in keyword_counts.items() if count > 1]
    assert not keyword_duplicates



from datetime import UTC, datetime

import numpy as np
from astropy.nddata import StdDevUncertainty
from ndcube import NDCube
from prefect import flow, get_run_logger

from punchbowl.data import NormalizedMetadata, get_base_file_name
from punchbowl.data.meta import set_spacecraft_location_to_earth
from punchbowl.data.wcs import load_quickpunch_mosaic_wcs, load_quickpunch_nfi_wcs
from punchbowl.level2.merge import merge_many_clear_task
from punchbowl.level2.resample import reproject_many_flow
from punchbowl.util import average_datetime, load_image_task, output_image_task

ORDER_QP = ["CR1", "CR2", "CR3", "CR4"]

@flow(validate_parameters=False)
def levelq_core_flow(data_list: list[str] | list[NDCube],
                     output_filename: list[str] | None = None) -> list[NDCube]:
    """Level quickPUNCH core flow."""
    logger = get_run_logger()
    logger.info("beginning level quickPUNCH core flow")

    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]

    if data_list:
        ordered_data_list: list[NDCube | None] = [None for _ in range(len(ORDER_QP))]
        for i, order_element in enumerate(ORDER_QP):
            for data_element in data_list:
                typecode = data_element.meta["TYPECODE"].value
                obscode = data_element.meta["OBSCODE"].value
                if typecode == order_element[:2] and obscode == order_element[2]:
                    ordered_data_list[i] = data_element
        logger.info("Ordered files are "
                    f"{[get_base_file_name(cube) if cube is not None else None for cube in ordered_data_list]}")

        quickpunch_mosaic_wcs, quickpunch_mosaic_shape = load_quickpunch_mosaic_wcs()
        quickpunch_nfi_wcs, quickpunch_nfi_shape = load_quickpunch_nfi_wcs()

        data_list_mosaic = reproject_many_flow(ordered_data_list, quickpunch_mosaic_wcs, quickpunch_mosaic_shape)
        output_dateobs = average_datetime([d.meta.datetime for d in data_list_mosaic]).isoformat()
        output_datebeg = min([d.meta.datetime for d in data_list_mosaic]).isoformat()
        output_dateend = max([d.meta.datetime for d in data_list_mosaic]).isoformat()

        data_list_nfi = reproject_many_flow(ordered_data_list[-1:], quickpunch_nfi_wcs, quickpunch_nfi_shape)
        output_dateobs_nfi = average_datetime([d.meta.datetime for d in data_list_nfi]).isoformat()
        output_datebeg_nfi = min([d.meta.datetime for d in data_list_nfi]).isoformat()
        output_dateend_nfi = max([d.meta.datetime for d in data_list_nfi]).isoformat()

        output_data_mosaic = merge_many_clear_task(data_list_mosaic, quickpunch_mosaic_wcs, level="Q")

        output_data_mosaic.meta["DATE"] = datetime.now(UTC).isoformat()
        output_data_mosaic.meta["DATE-AVG"] = output_dateobs
        output_data_mosaic.meta["DATE-OBS"] = output_dateobs
        output_data_mosaic.meta["DATE-BEG"] = output_datebeg
        output_data_mosaic.meta["DATE-END"] = output_dateend
        output_data_mosaic = set_spacecraft_location_to_earth(output_data_mosaic)

        output_meta_nfi = NormalizedMetadata.load_template("CNN", "Q")
        output_meta_nfi["DATE-OBS"] = data_list_nfi[0].meta["DATE-OBS"].value
        output_data_nfi =  NDCube(
            data=data_list_nfi[0].data,
            uncertainty=StdDevUncertainty(data_list_nfi[0].uncertainty.array),
            wcs=quickpunch_nfi_wcs,
            meta=output_meta_nfi,
            )
        output_data_nfi.meta["DATE"] = datetime.now(UTC).isoformat()
        output_data_nfi.meta["DATE-AVG"] = output_dateobs_nfi
        output_data_nfi.meta["DATE-OBS"] = output_dateobs_nfi
        output_data_nfi.meta["DATE-BEG"] = output_datebeg_nfi
        output_data_nfi.meta["DATE-END"] = output_dateend_nfi
        output_data_nfi = set_spacecraft_location_to_earth(output_data_nfi)
    else:
        output_dateobs = datetime.now(UTC).isoformat()
        output_datebeg = output_dateobs
        output_dateend = output_datebeg

        quickpunch_mosaic_wcs, quickpunch_mosaic_shape = load_quickpunch_mosaic_wcs()
        quickpunch_nfi_wcs, quickpunch_nfi_shape = load_quickpunch_nfi_wcs()
        output_data_mosaic = NDCube(
            data=np.zeros(quickpunch_mosaic_shape),
            uncertainty=StdDevUncertainty(np.zeros(quickpunch_mosaic_shape)),
            wcs=quickpunch_mosaic_wcs,
            meta=NormalizedMetadata.load_template("CTM", "Q"),
        )
        output_data_mosaic.meta["DATE"] = datetime.now(UTC).isoformat()
        output_data_mosaic.meta["DATE-AVG"] = output_dateobs
        output_data_mosaic.meta["DATE-OBS"] = output_dateobs
        output_data_mosaic.meta["DATE-BEG"] = output_datebeg
        output_data_mosaic.meta["DATE-END"] = output_dateend
        output_data_mosaic = set_spacecraft_location_to_earth(output_data_mosaic)

        output_data_nfi = NDCube(
            data=np.zeros(quickpunch_nfi_shape),
            uncertainty=StdDevUncertainty(np.zeros(quickpunch_nfi_shape)),
            wcs=quickpunch_nfi_wcs,
            meta=NormalizedMetadata.load_template("CNN", "Q"),
        )

        output_data_nfi.meta["DATE"] = datetime.now(UTC).isoformat()
        output_data_nfi.meta["DATE-AVG"] = output_dateobs
        output_data_nfi.meta["DATE-OBS"] = output_dateobs
        output_data_nfi.meta["DATE-BEG"] = output_dateobs
        output_data_nfi.meta["DATE-END"] = output_dateobs
        output_data_nfi = set_spacecraft_location_to_earth(output_data_nfi)


    if output_filename is not None:
        output_image_task(output_data_mosaic, output_filename[0])
        output_image_task(output_data_nfi, output_filename[1])

    logger.info("ending level quickPUNCH core flow")
    return [output_data_mosaic, output_data_nfi]



import os
from pathlib import Path
from collections.abc import Callable

import numpy as np
import reproject
from astropy.wcs import WCS
from ndcube import NDCube
from prefect import get_run_logger
from regularizepsf import ArrayPSF, ArrayPSFBuilder, ArrayPSFTransform, simple_functional_psf, varied_functional_psf
from regularizepsf.util import calculate_covering

from punchbowl.data.punch_io import load_ndcube_from_fits
from punchbowl.prefect import punch_task


def build_psf_transform(image_paths: list[str] | list[Path],
                        alpha: float = 0.7,
                        epsilon: float = 0.515,
                        target_sigma: float = 3.3,
                        psf_size: int = 64) -> ArrayPSFTransform:
    """
    Build the PSF transform for Level 1 processing from a list of images.

    Parameters
    ----------
    image_paths : List[str] | List[Path]
        images to use in building the PSF model
    alpha : float
        alpha parameter of the PSF transform, see Coma Off It paper
    epsilon : float
        epsilon parameter of the PSF transform, see Coma Off It paper
    target_sigma : float
        the target sigma of the PSF model in pixels
    psf_size : int
        size of the grid in the PSF model in pixels

    Returns
    -------
    ArrayPSFTransform
        the PSF transform that corresponds to the input images

    """
    b = ArrayPSFBuilder(psf_size)
    model, counts = b.build(image_paths)

    first_cube = load_ndcube_from_fits(image_paths[0], key="A")

    corrected_psf = generate_projected_psf(
            first_cube.wcs, psf_width=psf_size, star_gaussian_sigma=target_sigma)
    return ArrayPSFTransform.construct(model, corrected_psf, alpha=alpha, epsilon=epsilon)

def generate_projected_psf(
        source_wcs: WCS,
        psf_width: int = 64,
        star_gaussian_sigma: float = 3.3 / 2.355) -> ArrayPSF:
    """Create a varying PSF reflecting how a true circle looks in the mosaic image projection."""
    # Create a Gaussian star
    coords = np.arange(psf_width) - psf_width / 2 + .5
    xx, yy = np.meshgrid(coords, coords)
    perfect_star = np.exp(-(xx ** 2 + yy ** 2) / (2 * star_gaussian_sigma ** 2))

    star_wcs = WCS(naxis=2)
    star_wcs.wcs.ctype = "RA---ARC", "DEC--ARC"
    star_wcs.wcs.crpix = psf_width / 2 + .5, psf_width / 2 + .5
    star_wcs.wcs.cdelt = source_wcs.wcs.cdelt

    @simple_functional_psf
    def projected_psf(row: np.ndarray,  # noqa: ARG001
                      col: np.ndarray,  # noqa: ARG001
                      i: int = 0,
                      j: int = 0) -> np.ndarray:
        # Work out the center of this PSF patch
        ic = i + psf_width / 2 - .5
        jc = j + psf_width / 2 - .5
        ra, dec = source_wcs.array_index_to_world_values(ic, jc)

        # Create a WCS that places a star at that exact location
        swcs = star_wcs.deepcopy()
        swcs.wcs.crval = ra, dec

        # Project the star into this patch of the full image, telling us what a round
        # star looks like in this projection, distortion, etc.
        psf = reproject.reproject_adaptive(
            (perfect_star, swcs),
            source_wcs[i:i + psf_width, j:j + psf_width],
            (psf_width, psf_width),
            roundtrip_coords=False, return_footprint=False,
            boundary_mode="grid-constant", boundary_fill_value=0)
        return psf / np.sum(psf)

    @varied_functional_psf(projected_psf)
    def varying_projected_psf(row: int, col: int) -> dict:
        # row and col seem to be the upper-left corner of the image patch we're to describe
        return {"i": row, "j": col}

    coords = calculate_covering(source_wcs.array_shape, psf_width)
    return varying_projected_psf.as_array_psf(coords, psf_width)


def correct_psf(
    data: NDCube,
    psf_transform: ArrayPSFTransform,
    max_workers: int | None = None,
) -> NDCube:
    """Correct PSF."""
    new_data = psf_transform.apply(data.data, workers=max_workers)

    data.data[...] = new_data[...]
    # TODO: uncertainty propagation
    return data

@punch_task
def correct_psf_task(
    data_object: NDCube,
    model_path: str | Callable | None = None,
    max_workers: int | None = None,
) -> NDCube:
    """
    Prefect Task to correct the PSF of an image.

    Parameters
    ----------
    data_object : NDCube
        data to operate on
    model_path : str
        path to the PSF model to use in the correction
    max_workers : int
        the maximum number of worker threads to use

    Returns
    -------
    NDCube
        modified version of the input with the PSF corrected

    """
    if model_path is not None:
        if isinstance(model_path, Callable):
            corrector, model_path = model_path()
        else:
            corrector = ArrayPSFTransform.load(Path(model_path))
        data_object = correct_psf(data_object, corrector, max_workers)
        data_object.meta.history.add_now("LEVEL1-correct_psf",
                                         f"PSF corrected with {os.path.basename(model_path)} model")
        data_object.meta["CALPSF"] = os.path.basename(model_path)
    else:
        data_object.meta.history.add_now("LEVEL1-correct_psf", "Empty model path so no correction applied")
        logger = get_run_logger()
        logger.info("No model path so PSF correction is skipped")
    return data_object



import astropy.constants as const
import astropy.units as u
import numpy as np
import scipy.interpolate
from astropy.wcs import WCS
from numpy import ndarray

MSB = u.def_unit("MSB", 2.0090000E7 * u.W / u.m ** 2 / u.sr)


def calculate_image_pixel_area(wcs: WCS, data_shape: tuple[int, int], stride: int = 1) -> u.sr:
    """Calculate the sky area of every pixel in an image according to its WCS."""
    xx, yy = np.meshgrid(np.arange(0, data_shape[1] + stride, stride)-0.5,
                         np.arange(0, data_shape[0] + stride, stride)-0.5)
    coords = wcs.pixel_to_world(xx, yy)
    dx = coords[:, 1:].separation(coords[:, :-1]).to(u.deg)[:-1]
    dy = coords[1:, :].separation(coords[:-1, :]).to(u.deg)[:, :-1]
    area = dx * dy
    if stride > 1:
        area /= stride**2
        # Set these areas to be valid at the center of the region they span
        gridx, gridy = xx[0], yy[:, 0]
        gridx = (gridx[:-1] + gridx[1:]) / 2
        gridy = (gridy[:-1] + gridy[1:]) / 2
        # With fill_value=None, this will extrapolate to the edge pixels
        interpolator = scipy.interpolate.RegularGridInterpolator(
            (gridy, gridx), area, bounds_error=False, fill_value=None)
        xx, yy = np.meshgrid(np.arange(0, data_shape[1], 1),
                             np.arange(0, data_shape[0], 1))
        unit = area.unit
        area = interpolator(np.stack((yy, xx), axis=-1))
        area <<= unit
    return area

def msb_to_dn(data: ndarray,
              data_wcs: WCS,
              gain_left: float = 4.9 * u.photon / u.DN,
              gain_right: float = 4.9 * u.photon / u.DN,
              wavelength: float = 530. * u.nm,
              exposure: float = 49 * u.s,
              aperture: float = 49.57 * u.mm**2,
              pixel_area_stride: int = 1,
              ) -> ndarray:
    """Convert mean solar brightness to DNs."""
    energy_per_photon = (const.h * const.c / wavelength).to(u.J) / u.photon
    photon_flux = MSB / energy_per_photon
    pixel_scale = calculate_image_pixel_area(data_wcs, data.shape, pixel_area_stride).to(u.sr) / u.pixel
    photon_count = (photon_flux * exposure * aperture * pixel_scale * u.pixel).decompose()
    gain = split_ccd_array(data.shape, gain_left, gain_right)
    return data * photon_count / gain


def dn_to_msb(data: ndarray,
              data_wcs: WCS,
              gain_left: float = 4.9 * u.photon / u.DN,
              gain_right: float = 4.9 * u.photon / u.DN,
              wavelength: float = 530. * u.nm,
              exposure: float = 49 * u.s,
              aperture: float = 34 * u.mm**2,
              pixel_area_stride: int = 1,
              pixel_scale: u.Quantity = None,
              ) -> ndarray:
    """Convert DN to mean solar brightness."""
    energy_per_photon = (const.h * const.c / wavelength).to(u.J) / u.photon
    photon_flux = MSB / energy_per_photon
    if pixel_scale is None:
        pixel_scale = calculate_image_pixel_area(data_wcs, data.shape, pixel_area_stride).to(u.sr) / u.pixel
    photon_count = (photon_flux * exposure * aperture * pixel_scale * u.pixel).decompose()
    gain = split_ccd_array(data.shape, gain_left, gain_right)
    return data * gain / photon_count


def split_ccd_array(shape:tuple, value_left:float, value_right:float) -> ndarray:
    """Generate parameters across CCD halves."""
    array = np.zeros(shape)
    array[:,:shape[1]//2] = value_left
    array[:,shape[1]//2:] = value_right
    # TODO - Will need to either transpose output array or rename left/right across punchbowl
    return array.T



import os
from datetime import UTC, datetime

from ndcube import NDCube
from prefect import get_run_logger

from punchbowl.data.meta import NormalizedMetadata, set_spacecraft_location_to_earth
from punchbowl.level3.f_corona_model import subtract_f_corona_background_task
from punchbowl.level3.low_noise import create_low_noise_task
from punchbowl.level3.polarization import convert_polarization
from punchbowl.level3.stellar import subtract_starfield_background_task
from punchbowl.level3.velocity import plot_flow_map, track_velocity
from punchbowl.prefect import punch_flow
from punchbowl.util import load_image_task, output_image_task


@punch_flow
def level3_PIM_flow(data_list: list[str] | list[NDCube],  # noqa: N802
                     before_f_corona_model_path: str,
                     after_f_corona_model_path: str,
                     output_filename: str | None = None,
                     reference_time: datetime | None = None) -> list[NDCube]:  # noqa: ARG001
    """Level 3 PIM/CIM flow."""
    logger = get_run_logger()

    logger.info("beginning level 3 PIM/CIM flow")
    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]
    new_type = "CIM" if data_list[0].meta["TYPECODE"].value == "CT" else "PIM"

    data_list = [subtract_f_corona_background_task(d,
                                                   before_f_corona_model_path,
                                                   after_f_corona_model_path) for d in data_list]
    output_meta = NormalizedMetadata.load_template(new_type, "3")

    out_list = [NDCube(data=d.data, wcs=d.wcs, meta=output_meta) for d in data_list]
    for o, d in zip(out_list, data_list, strict=True):
        o.meta["DATE"] = datetime.now(UTC).isoformat()
        o.meta["DATE-AVG"] = d.meta["DATE-AVG"].value
        o.meta["DATE-OBS"] = d.meta["DATE-OBS"].value
        o.meta["DATE-BEG"] = d.meta["DATE-BEG"].value
        o.meta["DATE-END"] = d.meta["DATE-END"].value
        o = set_spacecraft_location_to_earth(o)   # noqa: PLW2901

    logger.info("ending level 3 PIM/CIM flow")

    if output_filename is not None:
        output_image_task(out_list[0], output_filename)

    return out_list


@punch_flow
def level3_core_flow(data_list: list[str] | list[NDCube],
                     before_f_corona_model_path: str,
                     after_f_corona_model_path: str,
                     starfield_background_path: str | None,
                     output_filename: str | None = None) -> list[NDCube]:
    """Level 3 core flow."""
    logger = get_run_logger()

    logger.info("beginning level 3 core flow")
    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]
    data_list = [subtract_f_corona_background_task(d,
                                                   before_f_corona_model_path,
                                                   after_f_corona_model_path) for d in data_list]
    data_list = [subtract_starfield_background_task(d, starfield_background_path) for d in data_list]
    if data_list[0].meta["TYPECODE"].value == "PT":
        data_list = [convert_polarization(d) for d in data_list]
    logger.info("ending level 3 core flow")


    if output_filename is not None:
        output_image_task(data_list[0], output_filename)

    return data_list


@punch_flow
def generate_level3_low_noise_flow(data_list: list[str] | list[NDCube],
                                   output_filename: str | None = None) -> list[NDCube]:
    """Generate low noise products."""
    logger = get_run_logger()

    logger.info("Generating low noise products")
    data_list = [load_image_task(d) if isinstance(d, str) else d for d in data_list]
    low_noise_image = create_low_noise_task(data_list)

    if output_filename is not None:
        output_image_task(low_noise_image, output_filename)

    return [low_noise_image]


@punch_flow
def generate_level3_velocity_flow(data_list: list[str],
                                  output_filename: str | None = None) -> list[NDCube]:
    """Generate Level 3 velocity data product."""
    logger = get_run_logger()

    logger.info("Generating velocity data product")
    velocity_data, plot_parameters = track_velocity(data_list)

    if output_filename is not None:
        output_image_task(velocity_data, output_filename)
        plot_filename = f"{os.path.splitext(output_filename)[0]}.png"
        plot_flow_map(plot_filename, **plot_parameters)

    return [velocity_data]
